from numba import jit

import numpy
from math import sqrt

# from matplotlib import pyplot as plt
#
# %matplotlib inline
from numpy import linspace

from numba import guvectorize

ttol = 1e-10


@jit
def PhiFB(a, b):
    return a + b - sqrt(a ** 2 + b ** 2)


@jit
def smooth_FB(f, x, a, b):
    return PhiFB(-PhiFB(f, x - a), b - x)


@jit
def PhiFB_(a, b):
    sq = sqrt(a ** 2 + b ** 2)
    x = a + b - sq
    if sq != 0.0:
        dx_da = 1.0 - a / sq
        dx_db = 1.0 - b / sq
    else:
        dx_da = 0.0
        dx_db = 0.0
    return (x, dx_da, dx_db)


@jit
def smooth_FB_(f, x, a, b, df):
    # P1 = Phi(f, x-a)
    # P2 = Phi(-P1, b-x)
    # f>=0 x-a>=04
    v, dv_df, dv_dx = PhiFB_(f, x - a)
    P1, dP1_dx = (v, dv_df * df + dv_dx)
    v, dv_df, dv_dx = PhiFB_(-P1, b - x)
    P2, dP2_dx = (v, -dv_df * dP1_dx - dv_dx)
    # phi, d_phi_d_x
    return (P2, dP2_dx)


import numpy as np


@jit
def smooth(f, x, a, b, df):
    n_ms, N, n_x = f.shape
    V = np.zeros_like(f)
    dV = df.copy()
    for i in range(n_ms):
        for n in range(N):
            for j in range(n_x):
                v, dv = smooth_FB_(
                    f[i, n, j], x[i, n, j], a[i, n, j], b[i, n, j], df[i, n, j, j]
                )
                V[i, n, j] = v
                # dv[i,n,j] =
    return V


#


@jit
def smooth_(f, x, a, b):
    n_ms, N, n_x = f.shape
    V = np.zeros_like(f)
    for i in range(n_ms):
        for n in range(N):
            for j in range(n_x):
                v = smooth_FB(f[i, n, j], x[i, n, j], a[i, n, j], b[i, n, j])
                V[i, n, j] = v
                # dv[i,n,j] =
    return V


#

# @jit
# def smooth(f,x,a,b,df):
#     n_ms, N, n_x = f.shape
#     V = np.zeros_like(f)
#     dV = df.copy()
#     for i in range(n_ms):
#         for n in range(N):
#             for j in range(n_x):
#                 v,dv = smooth_FB_(f[i,n,j], x[i,n,j], a[i,n,j], b[i,n,j], df[i,n,j,j])
#                 V[i,n,j] = v
#                 # dv[i,n,j] =
#     return V


#
#
# import numpy as np
#
# xvec = numpy.linspace(-0.1,1.1,2000)
# f = lambda x: -(x-1.2)*0.2
# fvec = f(xvec)
# dfvec = xvec*0.0-0.2
#
# phi_vec = np.array( [smooth_FB(fvec[i], xvec[i], 0.0, 1.0, dfvec[i]) for i in range(len(xvec))] )
# phi_vec_1 = np.array( [smooth_FB_(fvec[i], xvec[i], 0.0, 1.0, dfvec[i])[0] for i in range(len(xvec))] )
# d_phi = np.array( [smooth_FB_(fvec[i], xvec[i], 0.0, 1.0, dfvec[i])[1] for i in range(len(xvec))] )
#
# emp_diff = (phi_vec[1:]-phi_vec[:-1])/(xvec[1:]-xvec[:-1])
#
#
# plt.subplot(211)
# plt.plot(xvec, fvec)
# plt.plot(xvec,phi_vec)
# plt.plot(xvec, xvec*0, linestyle='--')
# plt.subplot(212)
# plt.plot(xvec, d_phi)
# plt.plot(xvec[:-1], emp_diff)



from __future__ import print_function

old_print = print

from numpy import zeros_like, zeros
from numpy.linalg import solve

import numpy
from numba import guvectorize
from numba import double


def solve(m, sol):

    # seems to segfault on windows

    h, w = m.shape

    for y in range(0, h):
        maxrow = y
        for y2 in range(y + 1, h):  # Find max pivot
            if abs(m[y2, y]) > abs(m[maxrow, y]):
                maxrow = y2
        for y2 in range(0, w):
            t = m[y, y2]
            m[y, y2] = m[maxrow, y2]
            m[maxrow, y2] = t

        for y2 in range(y + 1, h):  # Eliminate column y
            c = m[y2, y] / m[y, y]
            for x in range(y, w):
                m[y2, x] -= m[y, x] * c

    for y in range(h - 1, 0 - 1, -1):  # Backsubstitute
        c = m[y, y]
        for y2 in range(0, y):
            for x in range(w - 1, y - 1, -1):
                m[y2, x] -= m[y, x] * m[y2, y] / c
        m[y, y] /= c
        for x in range(h, w):  # Normalize row y
            m[y, x] /= c

    for y in range(h):
        sol[y] = m[y, w - 1]


serial_solve_numba = guvectorize("void(f8[:,:], f8[:])", "(m,n)->(m)")(solve)


from numpy.linalg import solve as linalg_solve


def serial_solve(A, B, diagnose=True):

    if diagnose:

        sol = zeros_like(B)

        for i in range(sol.shape[0]):
            try:
                sol[i, :] = linalg_solve(A[i, :, :], B[i, :])
            except:
                # Should be a special type of exception
                a = Exception("Error solving point {}".format(i))
                a.x = B[i, :]
                a.J = A[i, :, :]
                a.i = i
                raise a

    else:
        M = numpy.concatenate([A, B[:, :, None]], axis=2)
        sol = numpy.zeros_like(B)
        serial_solve_numba(M, sol)

    return sol


#
#     return sol

import time


def newton(f, x, verbose=False, tol=1e-6, maxit=5, jactype="serial"):

    """Solve nonlinear system using safeguarded Newton iterations


    Parameters
    ----------

    Return
    ------
    """
    if verbose:
        print = lambda txt: old_print(txt)
    else:
        print = lambda txt: None

    it = 0
    error = 10
    converged = False
    maxbacksteps = 30

    x0 = x

    if jactype == "sparse":
        from scipy.sparse.linalg import spsolve as solve
    elif jactype == "full":
        from numpy.linalg import solve
    else:
        solve = serial_solve

    while it < maxit and not converged:

        [v, dv] = f(x)

        # TODO: rewrite starting here

        #        print("Time to evaluate {}".format(ss-tt)0)

        error_0 = abs(v).max()

        if error_0 < tol:

            if verbose:
                print(
                    "> System was solved after iteration {}. Residual={}".format(
                        it, error_0
                    )
                )
            converged = True

        else:

            it += 1

            dx = solve(dv, v)

            # norm_dx = abs(dx).max()

            for bck in range(maxbacksteps):
                xx = x - dx * (2 ** (-bck))
                vm = f(xx)[0]
                err = abs(vm).max()
                if err < error_0:
                    break

            x = xx

            if verbose:
                print("\t> {} | {} | {}".format(it, err, bck))

    if not converged:
        import warnings

        warnings.warn("Did not converge")
    return [x, it]


serial_newton = newton

from numpy import sqrt, finfo, inf

from numpy import isinf, newaxis, diag, zeros


def SerialDifferentiableFunction(f, epsilon=1e-8):
    def df(x):

        v0 = f(x)

        N = v0.shape[0]
        n_v = v0.shape[1]
        assert x.shape[0] == N
        n_x = x.shape[1]

        dv = zeros((N, n_v, n_x))

        for i in range(n_x):

            xi = x.copy()
            xi[:, i] += epsilon

            vi = f(xi)

            dv[:, :, i] = (vi - v0) / epsilon

        return [v0, dv]

    return df


def test_serial_solve():

    N = 10

    import numpy

    A = numpy.random.random((N, 2, 2))

    B = numpy.random.random((N, 2))

    print(A)
    print(B)
    out = serial_solve(A, B)
    print("A")
    print(A)
    print("B")
    print(B)
    print("out")
    print(out)

    import numpy.linalg

    print(numpy.linalg.solve(A[0, :, :], B[0, :]))


if __name__ == "__main__":

    test_serial_solve()



"""
Discretization of continuous processes as markov chain
"""


import scipy as sp
import scipy.stats
import numpy as np
import math

# univariate tauchen code is taken from dpsolve ( https://bitbucket.org/stevejb/dpsolve )


def tauchen(N, mu, rho, sigma, m=2):
    """
    Approximate an AR1 process by a finite markov chain using Tauchen's method.

    :param N: scalar, number of nodes for Z
    :param mu: scalar, unconditional mean of process
    :param rho: scalar
    :param sigma: scalar, std. dev. of epsilons
    :param m: max +- std. devs.
    :returns: Z, N*1 vector, nodes for Z. Zprob, N*N matrix, transition probabilities

    SJB: This is a port of Martin Floden's 1996 Matlab code to implement Tauchen 1986 Economic Letters method The following comments are Floden's. Finds a Markov chain whose sample paths approximate those of the AR(1) process z(t+1) = (1-rho)*mu + rho * z(t) + eps(t+1) where eps are normal with stddev sigma.
    """
    Z = np.zeros((N, 1))
    Zprob = np.zeros((N, N))
    a = (1 - rho) * mu

    Z[-1] = m * math.sqrt(sigma ** 2 / (1 - (rho ** 2)))
    Z[0] = -1 * Z[-1]
    zstep = (Z[-1] - Z[0]) / (N - 1)

    for i in range(1, N):
        Z[i] = Z[0] + zstep * (i)

    Z = Z + a / (1 - rho)

    for j in range(0, N):
        for k in range(0, N):
            if k == 0:
                Zprob[j, k] = sp.stats.norm.cdf(
                    (Z[0] - a - rho * Z[j] + zstep / 2) / sigma
                )
            elif k == (N - 1):
                Zprob[j, k] = 1 - sp.stats.norm.cdf(
                    (Z[-1] - a - rho * Z[j] - zstep / 2) / sigma
                )
            else:
                up = sp.stats.norm.cdf((Z[k] - a - rho * Z[j] + zstep / 2) / sigma)
                down = sp.stats.norm.cdf((Z[k] - a - rho * Z[j] - zstep / 2) / sigma)
                Zprob[j, k] = up - down

    return (Z, Zprob)


def rouwenhorst(rho, sigma, N):
    """
    Approximate an AR1 process by a finite markov chain using Rouwenhorst's method.

    :param rho: autocorrelation of the AR1 process
    :param sigma: conditional standard deviation of the AR1 process
    :param N: number of states
    :return [nodes, P]: equally spaced nodes and transition matrix
    """

    from numpy import sqrt, linspace, array, zeros

    sigma = float(sigma)

    if N == 1:
        nodes = array([0.0])
        transitions = array([[1.0]])
        return [nodes, transitions]

    p = (rho + 1) / 2
    q = p
    nu = sqrt((N - 1) / (1 - rho ** 2)) * sigma

    nodes = linspace(-nu, nu, N)
    sig_a = sigma
    n = 1
    #    mat0 = array( [[1]] )
    mat0 = array([[p, 1 - p], [1 - q, q]])
    if N == 2:
        return [nodes, mat0]
    for n in range(3, N + 1):
        mat = zeros((n, n))
        mat_A = mat.copy()
        mat_B = mat.copy()
        mat_C = mat.copy()
        mat_D = mat.copy()
        mat_A[:-1, :-1] = mat0
        mat_B[:-1, 1:] = mat0
        mat_C[1:, :-1] = mat0
        mat_D[1:, 1:] = mat0

        mat0 = p * mat_A + (1 - p) * mat_B + (1 - q) * mat_C + q * mat_D
        mat0[1:-1, :] = mat0[1:-1, :] / 2
    P = mat0
    return [nodes, P]


def multidimensional_discretization(rho, sigma, N=3, method="rouwenhorst", m=2):
    """
    Discretize an VAR(1) into a markov chain. The autoregression matrix is supposed to be a scalar.

    :param rho:
    :param sigma:
    :param N:
    :param method:
    :param m:
    :return:
    """

    # rho is assumed to be a scalar
    # sigma is a positive symmetric matrix
    # N number of points in each non-degenerate dimension
    # m : standard deviations to approximate

    import scipy.linalg
    from itertools import product

    d = sigma.shape[1]

    sigma = sigma.copy()

    zero_columns = np.where(sigma.sum(axis=0) == 0)[0]
    for i in zero_columns:
        sigma[i, i] = 1

    L = scipy.linalg.cholesky(sigma)

    N = int(N)

    if method == "tauchen":
        [nodes_1d, probas_1d] = tauchen(N, 0, rho, 1, m=m)
    elif method == "rouwenhorst":
        [nodes_1d, probas_1d] = rouwenhorst(rho, 1, N)

    markov_nodes = np.array(list(product(*([nodes_1d] * d)))).T
    markov_indices = np.array(list(product(*([range(N)] * d)))).T

    markov_nodes = np.dot(L, markov_nodes)

    transition_matrix = 1
    for i in range(d):
        transition_matrix = np.kron(transition_matrix, probas_1d)

    markov_nodes = np.ascontiguousarray(markov_nodes.T)

    for i in zero_columns:
        markov_nodes[:, i] = 0

    return [markov_nodes, transition_matrix]


def tensor_markov(*args):
    """Computes the product of two independent markov chains.

    :param m1: a tuple containing the nodes and the transition matrix of the first chain
    :param m2: a tuple containing the nodes and the transition matrix of the second chain
    :return: a tuple containing the nodes and the transition matrix of the product chain
    """

    if len(args) > 2:

        m1 = args[0]
        m2 = args[1]
        tail = args[2:]

        prod = tensor_markov(m1, m2)
        return tensor_markov(prod, *tail)

    elif len(args) == 2:

        m1, m2 = args

        n1, t1 = m1
        n2, t2 = m2

        n1 = np.array(n1, dtype=float)
        n2 = np.array(n2, dtype=float)
        t1 = np.array(t1, dtype=float)
        t2 = np.array(t2, dtype=float)

        assert n1.shape[0] == t1.shape[0] == t1.shape[1]
        assert n2.shape[0] == t2.shape[0] == t2.shape[1]

        t = np.kron(t1, t2)

        p = t1.shape[0]
        q = t2.shape[0]

        np.tile(n2, (1, p))
        # n = np.row_stack([
        #     np.repeat(n1, q, axis=1),
        #     np.tile( n2, (1,p))
        # ])
        n = np.column_stack([np.repeat(n1, q, axis=0), np.tile(n2, (p, 1))])
        return n, t

    else:
        raise Exception(
            "Incorrect number of arguments. Expected at least 2. Found {}.".format(
                len(args)
            )
        )


#
# quantization_data = '/home/pablo/quantization_grids/'
#
#
# def quantization_nodes(N,sigma):
#     import numpy
#     import numpy.linalg
#     assert( len(sigma.shape) == 2 )
#     var = numpy.diag(sigma)
#     d = sigma.shape[0]
#     [w, x] = standard_quantization_weights(N,d )
#     A = numpy.linalg.cholesky(sigma)
#     x = numpy.dot(A, x)
#     return [x,w]
#
#
# def quantization_weights(N,sigma):
#     [x,w] = quantization_nodes(N,sigma)
#     return [w,x]
#
# def standard_quantization_weights(N,d):
#     filename = quantization_data + '{0}_{1}_nopti'.format(N,d)
#     import numpy
#
#     try:
#         G = numpy.loadtxt(filename)
#     except Exception as e:
#         raise e
#
#     w = G[:N,0]
#     x = G[:N,1:d+1]
#
#     s = numpy.dot(w, x)
#     x = x - numpy.outer(w,s)
#
#     return [w,x.T]

if __name__ == "__main__":

    [Z, Zprob] = tauchen(5, 0, 0.8, 0.1, 1.5)

    import numpy

    sigma = numpy.diag([0.1, 0.1]) ** 2
    sigma[0, 1] = numpy.sqrt(0.5 * sigma[0, 0] * sigma[1, 1])
    sigma[1, 0] = numpy.sqrt(0.5 * sigma[0, 0] * sigma[1, 1])
    rho = 0.9

    [nodes, transition] = multidimensional_discretization(
        rho, sigma, 2, method="rouwenhorst"
    )

    transition0 = transition.copy() * 0
    transition0[1, 1] = 1.0

    print(nodes.shape)

    print(transition.shape)

    [nodes, transition] = tensor_markov((nodes, transition0), (nodes, transition))

    print(nodes.shape)
    print(transition.shape)



import numpy
from numpy import array, zeros


def calibration_to_vector(symbols, calibration_dict):

    from dolang.triangular_solver import solve_triangular_system
    from numpy import nan

    sol = solve_triangular_system(calibration_dict)

    calibration = dict()
    for group in symbols:
        # t = numpy.array([sol.get(s, nan) for s in symbols[group]], dtype=float)
        t = numpy.array([sol.get(s, nan) for s in symbols[group]], dtype=float)
        calibration[group] = t

    return calibration


def calibration_to_dict(symbols, calib):

    if not isinstance(symbols, dict):
        symbols = symbols.symbols

    d = dict()
    for group, values in calib.items():
        if group == "covariances":
            continue
        syms = symbols[group]
        for i, s in enumerate(syms):
            d[s] = values[i]

    return d


from dolo.compiler.misc import calibration_to_dict

import copy

equivalent_symbols = dict(actions="controls")


class LoosyDict(dict):
    def __init__(self, **kwargs):

        kwargs = kwargs.copy()
        if "equivalences" in kwargs.keys():
            self.__equivalences__ = kwargs.pop("equivalences")
        else:
            self.__equivalences__ = dict()
        super().__init__(**kwargs)

    def __getitem__(self, p):

        if p in self.__equivalences__.keys():
            k = self.__equivalences__[p]
        else:
            k = p
        return super().__getitem__(k)


class CalibrationDict:

    # cb = CalibrationDict(symbols, calib)
    # calib['states'] -> array([ 1.        ,  9.35497829])
    # calib['states','controls'] - > [array([ 1.        ,  9.35497829]), array([ 0.23387446,  0.33      ])]
    # calib['z'] - > 1.0
    # calib['z','y'] -> [1.0, 0.99505814380953039]

    def __init__(self, symbols, calib, equivalences=equivalent_symbols):
        calib = copy.deepcopy(calib)
        for v in calib.values():
            v.setflags(write=False)
        self.symbols = symbols
        self.flat = calibration_to_dict(symbols, calib)
        self.grouped = LoosyDict(
            **{k: v for (k, v) in calib.items()}, equivalences=equivalences
        )

    def __getitem__(self, p):
        if isinstance(p, tuple):
            return [self[e] for e in p]
        if p in self.symbols.keys() or (p in self.grouped.__equivalences__.keys()):
            return self.grouped[p]
        else:
            return self.flat[p]


def allocating_function(inplace_function, size_output):
    def new_function(*args, **kwargs):
        val = numpy.zeros(size_output)
        nargs = args + (val,)
        inplace_function(*nargs)
        if "diff" in kwargs:
            return numdiff(new_function, args)
        return val

    return new_function


def numdiff(fun, args):
    """Vectorized numerical differentiation"""

    # vectorized version

    epsilon = 1e-8
    args = list(args)
    v0 = fun(*args)
    N = v0.shape[0]
    l_v = len(v0)
    dvs = []
    for i, a in enumerate(args):
        l_a = (a).shape[1]
        dv = numpy.zeros((N, l_v, l_a))
        nargs = list(args)  # .copy()
        for j in range(l_a):
            xx = args[i].copy()
            xx[:, j] += epsilon
            nargs[i] = xx
            dv[:, :, j] = (fun(*nargs) - v0) / epsilon
        dvs.append(dv)
    return [v0] + dvs



"""Do I need a docstring here ?"""

from .bruteforce_lib import *
from .invert import *


from dolo.numeric.decision_rule import DecisionRule
from dolo.misc.itprinter import IterationsPrinter
from numba import jit
import numpy
import time
import scipy.sparse.linalg

from operator import mul
from functools import reduce

from dolo.numeric.optimize.newton import SerialDifferentiableFunction


def prod(l):
    return reduce(mul, l)


from math import sqrt
from numba import jit
import time
from numpy import array, zeros

import time


@jit
def inplace(Phi, J):
    a, b, c, d, e = J.shape
    for i_a in range(a):
        for i_b in range(b):
            for i_c in range(c):
                for i_d in range(d):
                    for i_e in range(e):
                        J[i_a, i_b, i_c, i_d, i_e] *= Phi[i_a, i_c, i_d]


def smooth(res, dres, jres, dx, pos=1.0):
    from numpy import sqrt

    # jres is modified
    dinf = dx > 100000
    n_m, N, n_x = res.shape
    sq = sqrt(res ** 2 + (dx) ** 2)
    H = res + (dx) - sq
    Phi_a = 1 - res / sq
    Phi_b = 1 - (dx) / sq
    H[dinf] = res[dinf]
    Phi_a[dinf] = 1.0
    Phi_b[dinf] = 0.0
    H_x = Phi_a[:, :, :, None] * dres
    for i_x in range(n_x):
        H_x[:, :, i_x, i_x] += Phi_b[:, :, i_x] * pos
    # H_xt = Phi_a[:,None,:,:,None]*jres
    inplace(Phi_a, jres)
    return H, H_x, jres
    # return H, H_x, H_xt


def smooth_nodiff(res, dx):
    from numpy import sqrt

    n_m, N, n_x = res.shape
    dinf = dx > 100000
    sq = sqrt(res ** 2 + (dx) ** 2)
    H = res + (dx) - sq
    H[dinf] = res[dinf]
    return H


@jit
def ssmul(A, B):
    # simple serial_mult (matrix times vector)
    N, a, b = A.shape
    NN, b = B.shape
    O = numpy.zeros((N, a))
    for n in range(N):
        for k in range(a):
            for l in range(b):
                O[n, k] += A[n, k, l] * B[n, l]
    return O


@jit
def ssmul_inplace(A, B, O):
    # simple serial_mult (matrix times vector)
    N, a, b = A.shape
    NN, b = B.shape
    # O = numpy.zeros((N,a))
    for n in range(N):
        for k in range(a):
            for l in range(b):
                O[n, k] += A[n, k, l] * B[n, l]
    return O


# make parallel using guvectorize ?
def d_filt_dx(π, M_ij, S_ij, n_m, N, n_x, dumdr):
    # OK, so res is probably not what we need to filter here.
    # s sh
    n_m, n_im = M_ij.shape[:2]
    dumdr.set_values(π)
    i = 0
    j = 0
    for i in range(n_m):
        π[i, :, :] = 0
        for j in range(n_im):
            A = M_ij[i, j, :, :, :]
            B = dumdr.eval_ijs(i, j, S_ij[i, j, :, :])
            π[i, :, :] += ssmul(A, B)
    return π


from scipy.sparse.linalg import LinearOperator


class Operator(LinearOperator):

    """Special Linear Operator"""

    def __init__(self, M_ij, S_ij, dumdr):
        self.M_ij = M_ij
        self.S_ij = S_ij
        self.n_m = M_ij.shape[0]
        self.N = M_ij.shape[2]
        self.n_x = M_ij.shape[3]
        self.dumdr = dumdr
        self.dtype = numpy.dtype("float64")
        self.counter = 0
        self.addid = False

    @property
    def shape(self):
        nn = self.n_m * self.N * self.n_x
        return (nn, nn)

    def _matvec(self, x):
        self.counter += 1
        xx = x.reshape((self.n_m, self.N, self.n_x))
        yy = self.apply(xx)
        if self.addid:
            yy = xx - yy  # preconditioned system
        return yy.ravel()

    def apply(self, π, inplace=False):
        M_ij = self.M_ij
        S_ij = self.S_ij
        n_m = self.n_m
        N = self.N
        n_x = self.n_x
        dumdr = self.dumdr
        if not inplace:
            π = π.copy()
        return d_filt_dx(π, M_ij, S_ij, n_m, N, n_x, dumdr)

    def as_matrix(self):

        arg = np.zeros((self.n_m, self.N, self.n_x))
        larg = arg.ravel()
        N = len(larg)
        J = numpy.zeros((N, N))
        for i in range(N):
            if i > 0:
                larg[i - 1] = 0.0
            larg[i] = 1.0
            J[:, i] = self.apply(arg).ravel()
        return J


def invert_jac(res, dres, jres, fut_S, dumdr, tol=1e-10, maxit=1000, verbose=False):

    n_m = res.shape[0]
    N = res.shape[1]
    n_x = res.shape[2]

    err0 = 0.0
    ddx = solve_gu(dres.copy(), res.copy())

    lam = -1.0
    lam_max = -1.0
    err_0 = abs(ddx).max()

    tot = ddx.copy()
    if verbose:
        print("Starting inversion")
    for nn in range(maxit):
        # operations are made in place in ddx
        ddx = d_filt_dx(ddx, jres, fut_S, n_m, N, n_x, dumdr)
        err = abs(ddx).max()
        lam = err / err_0
        lam_max = max(lam_max, lam)
        if verbose:
            print("- {} | {} | {}".format(err, lam, lam_max))
        tot += ddx
        err_0 = err
        if err < tol:
            break

    # tot += ddx*lam/(1-lam)
    return tot, nn, lam


def radius_jac(res, dres, jres, fut_S, dumdr, tol=1e-10, maxit=1000, verbose=False):

    from numpy import sqrt

    n_m = res.shape[0]
    N = res.shape[1]
    n_x = res.shape[2]

    err0 = 0.0

    norm2 = lambda m: sqrt((m ** 2).sum())

    import numpy.random

    π = (numpy.random.random(res.shape) * 2 - 1) * 1
    π /= norm2(π)

    verbose = True
    lam = 1.0
    lam_max = 0.0

    lambdas = []
    if verbose:
        print("Starting inversion")
    for nn in range(maxit):
        # operations are made in place in ddx
        # π = (numpy.random.random(res.shape)*2-1)*1
        # π /= norm2(π)
        π[:, :, :] /= lam
        π = d_filt_dx(π, jres, fut_S, n_m, N, n_x, dumdr)
        lam = norm2(π)
        lam_max = max(lam_max, lam)
        if verbose:
            print("- {} | {}".format(lam, lam_max))
        lambdas.append(lam)
    return (lam, lam_max, lambdas)


from dolo import dprint
from .results import AlgoResult, ImprovedTimeIterationResult


def improved_time_iteration(
    model,
    method="jac",
    dr0=None,
    dprocess=None,
    interp_method="cubic",
    mu=2,
    maxbsteps=10,
    verbose=False,
    tol=1e-8,
    smaxit=500,
    maxit=1000,
    complementarities=True,
    compute_radius=False,
    invmethod="iti",
    details=True,
):
    def vprint(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    itprint = IterationsPrinter(
        ("N", int),
        ("f_x", float),
        ("d_x", float),
        ("Time_residuals", float),
        ("Time_inversion", float),
        ("Time_search", float),
        ("Lambda_0", float),
        ("N_invert", int),
        ("N_search", int),
        verbose=verbose,
    )
    itprint.print_header("Start Improved Time Iterations.")

    f = model.functions["arbitrage"]
    g = model.functions["transition"]
    x_lb = model.functions["arbitrage_lb"]
    x_ub = model.functions["arbitrage_ub"]

    parms = model.calibration["parameters"]

    grid, dprocess_ = model.discretize()

    if dprocess is None:
        dprocess = dprocess_

    endo_grid = grid["endo"]
    exo_grid = grid["exo"]

    n_m = max(dprocess.n_nodes, 1)
    n_s = len(model.symbols["states"])

    if interp_method in ("cubic", "linear"):
        ddr = DecisionRule(
            exo_grid, endo_grid, dprocess=dprocess, interp_method=interp_method
        )
        ddr_filt = DecisionRule(
            exo_grid, endo_grid, dprocess=dprocess, interp_method=interp_method
        )
    else:
        raise Exception("Unsupported interpolation method.")

    # s = ddr.endo_grid
    s = endo_grid.nodes
    N = s.shape[0]
    n_x = len(model.symbols["controls"])
    x0 = (
        model.calibration["controls"][
            None,
            None,
        ]
        .repeat(n_m, axis=0)
        .repeat(N, axis=1)
    )

    if dr0 is not None:
        for i_m in range(n_m):
            x0[i_m, :, :] = dr0.eval_is(i_m, s)
    ddr.set_values(x0)

    steps = 0.5 ** numpy.arange(maxbsteps)

    lb = x0.copy()
    ub = x0.copy()
    for i_m in range(n_m):
        m = dprocess.node(i_m)
        lb[i_m, :] = x_lb(m, s, parms)
        ub[i_m, :] = x_ub(m, s, parms)

    x = x0

    # both affect the precision

    ddr.set_values(x)

    ## memory allocation

    n_im = dprocess.n_inodes(0)  # we assume it is constant for now

    jres = numpy.zeros((n_m, n_im, N, n_x, n_x))
    S_ij = numpy.zeros((n_m, n_im, N, n_s))

    for it in range(maxit):

        jres[...] = 0.0
        S_ij[...] = 0.0

        t1 = time.time()

        # compute derivatives and residuals:
        # res: residuals
        # dres: derivatives w.r.t. x
        # jres: derivatives w.r.t. ~x
        # fut_S: future states

        ddr.set_values(x)
        #
        # ub[ub>100000] = 100000
        # lb[lb<-100000] = -100000
        #
        # sh_x = x.shape
        # rr =euler_residuals(f,g,s,x,ddr,dp,parms, diff=False, with_jres=False,set_dr=True)
        # print(rr.shape)
        #
        # from iti.fb import smooth_
        # jj = smooth_(rr, x, lb, ub)
        #
        # print("Errors with complementarities")
        # print(abs(jj.max()))
        #
        # exit()
        #

        from dolo.numeric.optimize.newton import SerialDifferentiableFunction

        sh_x = x.shape
        ff = SerialDifferentiableFunction(
            lambda u: euler_residuals(
                f,
                g,
                s,
                u.reshape(sh_x),
                ddr,
                dprocess,
                parms,
                diff=False,
                with_jres=False,
                set_dr=False,
            ).reshape((-1, sh_x[2]))
        )
        res, dres = ff(x.reshape((-1, sh_x[2])))
        res = res.reshape(sh_x)
        dres = dres.reshape((sh_x[0], sh_x[1], sh_x[2], sh_x[2]))
        junk, jres, fut_S = euler_residuals(
            f,
            g,
            s,
            x,
            ddr,
            dprocess,
            parms,
            diff=False,
            with_jres=True,
            set_dr=False,
            jres=jres,
            S_ij=S_ij,
        )

        # if there are complementerities, we modify derivatives
        if complementarities:
            res, dres, jres = smooth(res, dres, jres, x - lb)
            res[...] *= -1
            dres[...] *= -1
            jres[...] *= -1
            res, dres, jres = smooth(res, dres, jres, ub - x, pos=-1.0)
            res[...] *= -1
            dres[...] *= -1
            jres[...] *= -1

        err_0 = abs(res).max()

        # premultiply by A
        jres[...] *= -1.0

        for i_m in range(n_m):
            for j_m in range(n_im):
                M = jres[i_m, j_m, :, :, :]
                X = dres[i_m, :, :, :].copy()
                sol = solve_tensor(X, M)

        t2 = time.time()

        # new version
        if invmethod == "gmres":
            ddx = solve_gu(dres.copy(), res.copy())
            L = Operator(jres, fut_S, ddr_filt)
            n0 = L.counter
            L.addid = True
            ttol = err_0 / 100
            sol = scipy.sparse.linalg.gmres(
                L, ddx.ravel(), tol=ttol
            )  # , maxiter=1, restart=smaxit)
            lam0 = 0.01
            nn = L.counter - n0
            tot = sol[0].reshape(ddx.shape)
        else:
            # compute inversion
            tot, nn, lam0 = invert_jac(
                res,
                dres,
                jres,
                fut_S,
                ddr_filt,
                tol=tol,
                maxit=smaxit,
                verbose=(verbose == "full"),
            )

        # lam, lam_max, lambdas = radius_jac(res,dres,jres,fut_S,tol=tol,maxit=1000,verbose=(verbose=='full'),filt=ddr_filt)

        # backsteps
        t3 = time.time()
        for i_bckstps, lam in enumerate(steps):
            new_x = x - tot * lam
            new_err = euler_residuals(
                f, g, s, new_x, ddr, dprocess, parms, diff=False, set_dr=True
            )

            if complementarities:
                new_err = smooth_nodiff(new_err, new_x - lb)
                new_err = smooth_nodiff(-new_err, ub - new_x)

            new_err = abs(new_err).max()
            if new_err < err_0:
                break

        err_2 = abs(tot).max()
        t4 = time.time()
        itprint.print_iteration(
            N=it,
            f_x=err_0,
            d_x=err_2,
            Time_residuals=t2 - t1,
            Time_inversion=t3 - t2,
            Time_search=t4 - t3,
            Lambda_0=lam0,
            N_invert=nn,
            N_search=i_bckstps,
        )
        if err_0 < tol:
            break

        x = new_x

    ddr.set_values(x)

    itprint.print_finished()

    # if compute_radius:
    #     return ddx,L
    #     lam, lam_max, lambdas = radius_jac(res,dres,jres,fut_S,ddr_filt,tol=tol,maxit=smaxit,verbose=(verbose=='full'))
    #     return ddr, lam, lam_max, lambdas
    # else:
    if not details:
        return ddr
    else:
        ddx = solve_gu(dres.copy(), res.copy())
        L = Operator(jres, fut_S, ddr_filt)

        if compute_radius:
            lam = scipy.sparse.linalg.eigs(L, k=1, return_eigenvectors=False)
            lam = abs(lam[0])
        else:
            lam = np.nan
        # lam, lam_max, lambdas = radius_jac(res,dres,jres,fut_S,ddr_filt,tol=tol,maxit=smaxit,verbose=(verbose=='full'))
        return ImprovedTimeIterationResult(
            ddr, it, err_0, err_2, err_0 < tol, complementarities, lam, None, L
        )


def euler_residuals(
    f,
    g,
    s,
    x,
    dr,
    dp,
    p_,
    diff=True,
    with_jres=False,
    set_dr=True,
    jres=None,
    S_ij=None,
):

    t1 = time.time()

    if set_dr:
        dr.set_values(x)

    N = s.shape[0]
    n_s = s.shape[1]
    n_x = x.shape[2]

    n_ms = max(dp.n_nodes, 1)  # number of markov states
    n_im = dp.n_inodes(0)

    res = numpy.zeros_like(x)

    if with_jres:
        if jres is None:
            jres = numpy.zeros((n_ms, n_im, N, n_x, n_x))
        if S_ij is None:
            S_ij = numpy.zeros((n_ms, n_im, N, n_s))

    for i_ms in range(n_ms):
        m_ = dp.node(i_ms)
        xm = x[i_ms, :, :]
        for I_ms in range(n_im):
            M_ = dp.inode(i_ms, I_ms)
            w = dp.iweight(i_ms, I_ms)
            S = g(m_, s, xm, M_, p_, diff=False)
            XM = dr.eval_ijs(i_ms, I_ms, S)
            if with_jres:
                ff = SerialDifferentiableFunction(
                    lambda u: f(m_, s, xm, M_, S, u, p_, diff=False)
                )
                rr, rr_XM = ff(XM)

                rr = f(m_, s, xm, M_, S, XM, p_, diff=False)
                jres[i_ms, I_ms, :, :, :] = w * rr_XM
                S_ij[i_ms, I_ms, :, :] = S
            else:
                rr = f(m_, s, xm, M_, S, XM, p_, diff=False)
            res[i_ms, :, :] += w * rr

    t2 = time.time()

    if with_jres:
        return res, jres, S_ij
    else:
        return res



from dolo.numeric.tensor import sdot, mdot

import numpy as np

TOL = 1e-10

# credits : second_order_solver is adapted from Sven Schreiber's port of Uhlig's Toolkit.
def second_order_solver(FF, GG, HH, eigmax=1.0 + 1e-6):

    # from scipy.linalg import qz
    from dolo.numeric.extern.qz import qzordered

    from numpy import (
        array,
        mat,
        c_,
        r_,
        eye,
        zeros,
        real_if_close,
        diag,
        allclose,
        where,
        diagflat,
    )
    from numpy.linalg import solve

    Psi_mat = array(FF)
    Gamma_mat = array(-GG)
    Theta_mat = array(-HH)
    m_states = FF.shape[0]

    Xi_mat = r_[
        c_[Gamma_mat, Theta_mat], c_[eye(m_states), zeros((m_states, m_states))]
    ]

    Delta_mat = r_[
        c_[Psi_mat, zeros((m_states, m_states))],
        c_[zeros((m_states, m_states)), eye(m_states)],
    ]

    [Delta_up, Xi_up, UUU, VVV, eigval] = qzordered(
        Delta_mat,
        Xi_mat,
    )

    VVVH = VVV.T
    VVV_2_1 = VVVH[m_states : 2 * m_states, :m_states]
    VVV_2_2 = VVVH[m_states : 2 * m_states, m_states : 2 * m_states]
    UUU_2_1 = UUU[m_states : 2 * m_states, :m_states]
    PP = -solve(VVV_2_1, VVV_2_2)

    # slightly different check than in the original toolkit:
    assert allclose(real_if_close(PP), PP.real)
    PP = PP.real

    return [eigval, PP]


def solve_sylvester_vectorized(*args):
    from numpy import kron
    from numpy.linalg import solve

    vec = lambda M: M.ravel()
    n = args[0][0].shape[0]
    q = args[0][1].shape[0]
    K = vec(args[-1])
    L = sum([kron(A, B.T) for (A, B) in args[:-1]])
    X = solve(L, -K)
    return X.reshape((n, q))


def solve_sylvester(A, B, C, D, Ainv=None, method="linear"):
    # Solves equation : A X + B X [C,...,C] + D = 0
    # where X is a multilinear function whose dimension is determined by D
    # inverse of A can be optionally specified as an argument

    n_d = D.ndim - 1
    n_v = C.shape[1]

    n_c = D.size // n_v ** n_d

    #    import dolo.config
    #    opts = dolo.config.use_engine
    #    if opts['sylvester']:
    #        DD = D.flatten().reshape( n_c, n_v**n_d)
    #        [err,XX] = dolo.config.engine.engine.feval(2,'gensylv',n_d,A,B,C,-DD)
    #        X = XX.reshape( (n_c,)+(n_v,)*(n_d))

    DD = D.reshape(n_c, n_v ** n_d)

    if n_d == 1:
        CC = C
    else:
        CC = np.kron(C, C)
    for i in range(n_d - 2):
        CC = np.kron(CC, C)

    if method == "linear":
        I = np.eye(CC.shape[0])
        XX = solve_sylvester_vectorized((A, I), (B, CC), DD)

    else:
        # we use slycot by default
        import slycot

        if Ainv != None:
            Q = sdot(Ainv, B)
            S = sdot(Ainv, DD)
        else:
            Q = np.linalg.solve(A, B)
            S = np.linalg.solve(A, DD)

        n = n_c
        m = n_v ** n_d

        XX = slycot.sb04qd(n, m, Q, CC, -S)

    X = XX.reshape((n_c,) + (n_v,) * (n_d))

    return X


class BKError(Exception):
    def __init__(self, type):
        self.type = type

    def __str__(self):
        return "Blanchard-Kahn error ({0})".format(self.type)



from dolo import *

from dolo.algos.value_iteration import constant_policy


def test_import():

    filenames = [
        "examples/models/rbc_iid.yaml",
        "examples/models/rbc_mc.yaml",
        "examples/models/rbc_ar1.yaml",
        "examples/models/rbc.yaml",
    ]

    for fname in filenames:

        model = yaml_import(fname, check=False)
        print(model)
        print("Exogenous shocks:")
        print(model.exogenous)
        print("Discretized shock:")
        print(model.exogenous.discretize())
        print(model.symbols)
        print(model.definitions)
        dprocess = model.exogenous.discretize()
        print(dprocess.n_nodes)
        print(dprocess.n_inodes(0))
        print(dprocess.inode(0, 0))
        print(dprocess.node(0))


def test_old_models():

    import os

    os.listdir("examples/models_")  # old models
    filenames = [
        f"examples/models_/{fname}" for fname in os.listdir("examples/models_")
    ]

    for fname in filenames:

        try:
            print(f"Importing: {fname}")
            model = yaml_import(fname, check=True)

        except Exception as e:
            print(fname)
            raise (e)
            # assert( not isinstance(e, Exception) )


if __name__ == "__main__":
    test_import()



"""Time Iteration Algorithm"""

import numpy
from dolo import dprint
from dolo.numeric.processes import DiscretizedIIDProcess
from dolo.numeric.decision_rule import DecisionRule
from dolo.numeric.grids import CartesianGrid


def residuals_simple(f, g, s, x, dr, dprocess, parms):

    N = s.shape[0]
    n_s = s.shape[1]

    res = numpy.zeros_like(x)

    for i_ms in range(dprocess.n_nodes):

        # solving on grid for markov index i_ms
        m = numpy.tile(dprocess.node(i_ms), (N, 1))
        xm = x[i_ms, :, :]

        for I_ms in range(dprocess.n_inodes(i_ms)):
            M = numpy.tile(dprocess.inode(i_ms, I_ms), (N, 1))
            prob = dprocess.iweight(i_ms, I_ms)
            S = g(m, s, xm, M, parms)
            XM = dr.eval_ijs(i_ms, I_ms, S)
            rr = f(m, s, xm, M, S, XM, parms)
            res[i_ms, :, :] += prob * rr

    return res


from .results import TimeIterationResult, AlgoResult


def time_iteration(
    model,
    dr0=None,
    with_complementarities=True,
    dprocess=None,
    verbose=True,
    maxit=1000,
    inner_maxit=10,
    tol=1e-6,
    hook=None,
    details=False,
    interp_method="cubic",
):

    """Finds a global solution for ``model`` using backward time-iteration.


    This algorithm iterates on the residuals of the arbitrage equations

    Parameters
    ----------
    model : Model
        model to be solved
    verbose : boolean
        if True, display iterations
    dr0 : decision rule
        initial guess for the decision rule
    with_complementarities : boolean (True)
        if False, complementarity conditions are ignored
    maxit: maximum number of iterations
    inner_maxit: maximum number of iteration for inner solver
    tol: tolerance criterium for successive approximations
    hook: Callable
        function to be called within each iteration, useful for debugging purposes


    Returns
    -------
    decision rule :
        approximated solution
    """

    from dolo import dprint

    def vprint(t):
        if verbose:
            print(t)

    grid, dprocess_ = model.discretize()

    if dprocess is None:
        dprocess = dprocess_

    n_ms = dprocess.n_nodes  # number of exogenous states
    n_mv = dprocess.n_inodes(0)  # this assume number of integration nodes is constant

    x0 = model.calibration["controls"]
    parms = model.calibration["parameters"]
    n_x = len(x0)
    n_s = len(model.symbols["states"])

    endo_grid = grid["endo"]
    exo_grid = grid["exo"]

    mdr = DecisionRule(
        exo_grid, endo_grid, dprocess=dprocess, interp_method=interp_method
    )

    s = mdr.endo_grid.nodes
    N = s.shape[0]

    controls_0 = numpy.zeros((n_ms, N, n_x))
    if dr0 is None:
        controls_0[:, :, :] = x0[None, None, :]
    else:
        if isinstance(dr0, AlgoResult):
            dr0 = dr0.dr
        try:
            for i_m in range(n_ms):
                controls_0[i_m, :, :] = dr0(i_m, s)
        except Exception:
            for i_m in range(n_ms):
                m = dprocess.node(i_m)
                controls_0[i_m, :, :] = dr0(m, s)

    f = model.functions["arbitrage"]
    g = model.functions["transition"]

    if "arbitrage_lb" in model.functions and with_complementarities == True:
        lb_fun = model.functions["arbitrage_lb"]
        ub_fun = model.functions["arbitrage_ub"]
        lb = numpy.zeros_like(controls_0) * numpy.nan
        ub = numpy.zeros_like(controls_0) * numpy.nan
        for i_m in range(n_ms):
            m = dprocess.node(i_m)[None, :]
            p = parms[None, :]
            m = numpy.repeat(m, N, axis=0)
            p = numpy.repeat(p, N, axis=0)

            lb[i_m, :, :] = lb_fun(m, s, p)
            ub[i_m, :, :] = ub_fun(m, s, p)

    else:
        with_complementarities = False

    sh_c = controls_0.shape

    controls_0 = controls_0.reshape((-1, n_x))

    from dolo.numeric.optimize.newton import newton, SerialDifferentiableFunction
    from dolo.numeric.optimize.ncpsolve import ncpsolve

    err = 10
    it = 0

    if with_complementarities:
        lb = lb.reshape((-1, n_x))
        ub = ub.reshape((-1, n_x))

    if verbose:
        headline = "|{0:^4} | {1:10} | {2:8} | {3:8} | {4:3} |".format(
            "N", " Error", "Gain", "Time", "nit"
        )
        stars = "-" * len(headline)
        print(stars)
        print(headline)
        print(stars)

    import time

    t1 = time.time()

    err_0 = numpy.nan

    verbit = verbose == "full"

    while err > tol and it < maxit:

        it += 1

        t_start = time.time()

        mdr.set_values(controls_0.reshape(sh_c))

        fn = lambda x: residuals_simple(
            f, g, s, x.reshape(sh_c), mdr, dprocess, parms
        ).reshape((-1, n_x))
        dfn = SerialDifferentiableFunction(fn)

        res = fn(controls_0)

        if hook:
            hook()

        if with_complementarities:
            [controls, nit] = ncpsolve(
                dfn, lb, ub, controls_0, verbose=verbit, maxit=inner_maxit
            )
        else:
            [controls, nit] = newton(dfn, controls_0, verbose=verbit, maxit=inner_maxit)

        err = abs(controls - controls_0).max()

        err_SA = err / err_0
        err_0 = err

        controls_0 = controls

        t_finish = time.time()
        elapsed = t_finish - t_start

        if verbose:
            print(
                "|{0:4} | {1:10.3e} | {2:8.3f} | {3:8.3f} | {4:3} |".format(
                    it, err, err_SA, elapsed, nit
                )
            )

    controls_0 = controls.reshape(sh_c)

    mdr.set_values(controls_0)

    t2 = time.time()

    if verbose:
        print(stars)
        print("Elapsed: {} seconds.".format(t2 - t1))
        print(stars)

    if not details:
        return mdr

    return TimeIterationResult(
        mdr,
        it,
        with_complementarities,
        dprocess,
        err < tol,  # x_converged: bool
        tol,  # x_tol
        err,  #: float
        None,  # log: object # TimeIterationLog
        None,  # trace: object #{Nothing,IterationTrace}
    )



import numpy
import pandas
import xarray as xr
import numpy as np
from dolo.numeric.optimize.ncpsolve import ncpsolve
from dolo.numeric.optimize.newton import newton as newton_solver
from dolo.numeric.optimize.newton import SerialDifferentiableFunction

## TODO: extend for mc process


def response(model, dr, varname, T=40, impulse: float = None):

    i_exo = model.symbols["exogenous"].index(varname)

    if impulse is None:
        try:
            impulse = numpy.sqrt(
                model.exogenous.Σ[i_exo, i_exo]
            )  # works only for IID/AR1
        except:
            impulse = numpy.sqrt(model.exogenous.σ)  # works only for IID/AR1

    e1 = numpy.zeros(len(model.symbols["exogenous"]))
    e1[i_exo] = impulse

    exogenous = model.exogenous
    print(exogenous)
    print(T, e1)
    m_simul = model.exogenous.response(T - 1, e1)  # this is an xarray T x V
    m_simul = m_simul.expand_dims("N")
    m_simul = m_simul.transpose("T", "N", "V").data

    sim = simulate(model, dr, N=1, T=T, driving_process=m_simul, stochastic=False)

    irf = sim.sel(N=0)

    return irf


def find_index(sim, values):
    sh = sim.shape
    N = sh[0]
    T = sh[1]
    indices = np.zeros((N, T), dtype=int)
    for n in range(N):
        for t in range(T):
            v = sim[n, t, :]
            ind = np.where((values == v[None, :]).all(axis=1))[0][0]
            indices[n, t] = ind
    return indices


from dolo.numeric.grids import CartesianGrid, UnstructuredGrid
from dolo.algos.results import AlgoResult


def simulate(
    model,
    dr,
    process=None,
    N=1,
    T=40,
    s0=None,
    i0=None,
    m0=None,
    driving_process=None,
    seed=42,
    stochastic=True,
):
    """
    Simulate a model using the specified decision rule.

    Parameters
    ----------

    model: Model

    dr: decision rule

    process:

    s0: ndarray
        initial state where all simulations start

    driving_process: ndarray
        realization of exogenous driving process (drawn randomly if None)

    N: int
        number of simulations
    T: int
        horizon for the simulations
    seed: int
        used to initialize the random number generator. Use it to replicate
        exact same results among simulations
    discard: boolean (False)
        if True, then all simulations containing at least one non finite value
        are discarded

    Returns
    -------
    xarray.DataArray:
        returns a ``T x N x n_v`` array where ``n_v``
           is the number of variables.
    """

    if isinstance(dr, AlgoResult):
        dr = dr.dr

    calib = model.calibration
    parms = numpy.array(calib["parameters"])

    if s0 is None:
        s0 = calib["states"]

    n_x = len(model.symbols["controls"])
    n_s = len(model.symbols["states"])

    s_simul = numpy.zeros((T, N, n_s))
    x_simul = numpy.zeros((T, N, n_x))

    s_simul[0, :, :] = s0[None, :]

    # are we simulating a markov chain or a continuous process ?
    if driving_process is not None:
        if len(driving_process.shape) == 3:
            m_simul = driving_process
            sim_type = "continuous"
            if m0 is None:
                m0 = model.calibration["exogenous"]
            x_simul[0, :, :] = dr.eval_ms(m0[None, :], s0[None, :])[0, :]
        elif len(driving_process.shape) == 2:
            i_simul = driving_process
            nodes = dr.exo_grid.nodes
            m_simul = nodes[i_simul]
            # inds = i_simul.ravel()
            # m_simul = np.reshape( np.concatenate( [nodes[i,:][None,:] for i in inds.ravel()], axis=0 ), inds.shape + (-1,) )
            sim_type = "discrete"
            x_simul[0, :, :] = dr.eval_is(i0, s0[None, :])[0, :]
        else:
            raise Exception("Incorrect specification of driving values.")
        m0 = m_simul[0, :, :]
    else:
        from dolo.numeric.processes import DiscreteProcess

        if process is None:
            if hasattr(dr, "dprocess") and hasattr(dr.dprocess, "simulate"):
                process = dr.dprocess
            else:
                process = model.exogenous

        # detect type of simulation
        if not isinstance(process, DiscreteProcess):
            sim_type = "continuous"
        else:
            sim_type = "discrete"

        if sim_type == "discrete":
            if i0 is None:
                i0 = 0
            dp = process
            m_simul = dp.simulate(N, T, i0=i0, stochastic=stochastic)
            i_simul = find_index(m_simul, dp.values)
            m0 = dp.node(i0)
            x0 = dr.eval_is(i0, s0[None, :])[0, :]
        else:
            m_simul = process.simulate(N, T, m0=m0, stochastic=stochastic)
            if isinstance(m_simul, xr.DataArray):
                m_simul = m_simul.data
            sim_type = "continuous"
            if m0 is None:
                m0 = model.calibration["exogenous"]
            x0 = dr.eval_ms(m0[None, :], s0[None, :])[0, :]
            x_simul[0, :, :] = x0[None, :]

    f = model.functions["arbitrage"]
    g = model.functions["transition"]

    numpy.random.seed(seed)

    mp = m0
    for i in range(T):
        m = m_simul[i, :, :]
        s = s_simul[i, :, :]
        if sim_type == "discrete":
            i_m = i_simul[i, :]
            xx = [
                dr.eval_is(i_m[ii], s[ii, :][None, :])[0, :] for ii in range(s.shape[0])
            ]
            x = np.row_stack(xx)
        else:
            x = dr.eval_ms(m, s)

        x_simul[i, :, :] = x

        ss = g(mp, s, x, m, parms)
        if i < T - 1:
            s_simul[i + 1, :, :] = ss
        mp = m

    if "auxiliary" not in model.functions:  # TODO: find a better test than this
        l = [s_simul, x_simul]
        varnames = model.symbols["states"] + model.symbols["controls"]
    else:
        aux = model.functions["auxiliary"]
        a_simul = aux(
            m_simul.reshape((N * T, -1)),
            s_simul.reshape((N * T, -1)),
            x_simul.reshape((N * T, -1)),
            parms,
        )
        a_simul = a_simul.reshape(T, N, -1)
        l = [m_simul, s_simul, x_simul, a_simul]
        varnames = (
            model.symbols["exogenous"]
            + model.symbols["states"]
            + model.symbols["controls"]
            + model.symbols["auxiliaries"]
        )

    simul = numpy.concatenate(l, axis=2)

    if sim_type == "discrete":
        varnames = ["_i_m"] + varnames
        simul = np.concatenate([i_simul[:, :, None], simul], axis=2)

    data = xr.DataArray(
        simul,
        dims=["T", "N", "V"],
        coords={"T": range(T), "N": range(N), "V": varnames},
    )

    return data


def tabulate(
    model, dr, state, bounds=None, n_steps=100, s0=None, i0=None, m0=None, **kwargs
):

    import numpy

    if isinstance(dr, AlgoResult):
        dr = dr.dr

    states_names = model.symbols["states"]
    controls_names = model.symbols["controls"]
    index = states_names.index(str(state))

    if bounds is None:
        try:
            endo_grid = dr.endo_grid
            bounds = [endo_grid.min[index], endo_grid.max[index]]
        except:
            domain = model.domain
            bounds = [domain.min[index], domain.max[index]]
        if bounds is None:
            raise Exception("No bounds provided for simulation or by model.")

    values = numpy.linspace(bounds[0], bounds[1], n_steps)

    if s0 is None:
        s0 = model.calibration["states"]

    svec = numpy.row_stack([s0] * n_steps)
    svec[:, index] = values

    try:
        dp = dr.dprocess
    except:
        dp = model.exogenous.discretize()

    if (i0 is None) and (m0 is None):
        from dolo.numeric.grids import UnstructuredGrid

        if isinstance(dp.grid, UnstructuredGrid):
            n_ms = dp.n_nodes
            [q, r] = divmod(n_ms, 2)
            i0 = q - 1 + r
        else:
            m0 = model.calibration["exogenous"]

    if i0 is not None:
        m = dp.node(i0)
        xvec = dr.eval_is(i0, svec)
    elif m0 is not None:
        m = m0
        xvec = dr.eval_ms(m0, svec)

    mm = numpy.row_stack([m] * n_steps)
    l = [mm, svec, xvec]

    series = (
        model.symbols["exogenous"] + model.symbols["states"] + model.symbols["controls"]
    )

    if "auxiliary" in model.functions:
        p = model.calibration["parameters"]
        pp = numpy.row_stack([p] * n_steps)
        avec = model.functions["auxiliary"](mm, svec, xvec, pp)
        l.append(avec)
        series.extend(model.symbols["auxiliaries"])

    import pandas

    tb = numpy.concatenate(l, axis=1)
    df = pandas.DataFrame(tb, columns=series)

    return df


def tabulate_2d(model, dr, states=None, i0=0, s0=None, n=[12, 13]):

    import numpy
    import xarray as xr

    if isinstance(dr, AlgoResult):
        dr = dr.dr

    if s0 is None:
        s0 = model.calibration["states"]
    if states is None:
        states = model.symbols["states"]
    assert len(states) == 2
    domain = model.get_domain()
    lps = [numpy.linspace(*domain[s], n[i]) for i, s in enumerate(states)]
    i_x = model.symbols["states"].index(states[0])
    i_y = model.symbols["states"].index(states[1])
    vals = []
    vstates = []
    s = s0.copy()
    for xx in lps[0]:
        vv = []
        s[i_x] = xx
        for yy in lps[1]:
            s[i_y] = yy
            x = dr.eval_is(i0, s)
            vv.append(numpy.concatenate([s, x]))
        vals.append(vv)
    vv = numpy.array(vals)
    controls = model.symbols["states"] + model.symbols["controls"]
    #     tab = xr.DataArray(vv, dims=[states[0], states[1], 'V'], coords=[lps[0], lps[1], 'V'])
    tab = xr.DataArray(
        vv,
        dims=[states[0], states[1], "V"],
        coords={states[0]: lps[0], states[1]: lps[1], "V": controls},
    )
    return tab


def plot3d(tab, varname):
    X = numpy.array(tab[tab.dims[0]])
    Y = numpy.array(tab[tab.dims[1]])
    Z = numpy.array(tab.loc[:, :, varname])
    data = [go.Surface(x=X, y=Y, z=Z)]
    layout = go.Layout(
        title="Equity",
        autosize=False,
        width=500,
        height=500,
        #         xaxis=go.XAxis(title=tab.dims[0]),
        #         yaxis={'title':tab.dims[1]},
        #         zaxis={'title':varname},
        xaxis=dict(
            title="x Axis",
            nticks=7,
            titlefont=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
        ),
        margin=dict(l=65, r=50, b=65, t=90),
    )
    fig = go.Figure(data=data, layout=layout)
    return iplot(fig, filename="graph_" + varname)


def plot_decision_rule(plot_controls=None, **kwargs):

    if isinstance(dr, AlgoResult):
        dr = dr.dr

    df = tabulate(dr, state, bounds=None, n_steps=100, s0=None, i0=None, m0=None)

    from matplotlib import pyplot

    if isinstance(plot_controls, str):
        cn = plot_controls
        pyplot.plot(values, df[cn], **kwargs)
    else:
        for cn in plot_controls:
            pyplot.plot(values, df[cn], label=cn, **kwargs)
        pyplot.legend()
    pyplot.xlabel("state = {} | mstate = {}".format(state, i0))



import time


class IterationsPrinter:
    def __init__(self, *knames, verbose=False):
        knames = dict(knames)
        names = []
        types = []
        fmts = []
        for k, v in knames.items():
            names.append(k)
            types.append(v)
            if v == int:
                mm = max([4, len(k)])
                # print(mm)
                fmts.append("{{:{}}}".format(mm))
            elif v == float:
                fmts.append("{:9.3e}")
        fmt_str = "| " + str.join(" | ", fmts) + " |"
        self.verbose = verbose
        self.names = names
        self.types = types
        self.fmts = fmts
        self.width = len(fmt_str.format(*[0 for i in self.names]))
        self.fmt_str = fmt_str
        self.t_start = time.time()

    def print_line(self):
        if not self.verbose:
            return
        print("-" * self.width)

    def print_header(self, msg=None):
        if not self.verbose:
            return
        self.print_line()
        if msg is not None:
            ll = "| " + msg
            print(ll + " " * (self.width - len(ll) - 1) + "|")
            self.print_line()
        title_str = ""
        for i, v in enumerate(self.types):
            k = self.names[i]
            if v == int:
                title_str += " {:4} |".format(k)
            elif v == float:
                title_str += " {:9} |".format(k)
        title_str = "|" + title_str
        print(title_str)
        self.print_line()

    def print_iteration(self, **args):
        if not self.verbose:
            return
        vals = [args[k] for k in self.names]
        print(self.fmt_str.format(*vals))

    def print_finished(self):
        if not self.verbose:
            return
        elapsed = time.time() - self.t_start
        line = "| Elapsed: {:.2f} seconds.".format(elapsed)
        self.print_line()
        print(line + " " * (self.width - len(line) - 1) + "|")
        self.print_line()
        print()



import unittest
import scipy
from scipy.integrate import quad
import numpy as np
from dolo.numeric.distribution import *
from dolo.numeric.processes import ConstantProcess

## Polynomial
def f(x):
    return x ** 2


def test_Mixture():
    μ_cons = 0.5
    σ = 0.1
    μ_norm = 0.2
    π = 0.4
    distr = Mixture(
        index=Bernouilli(π=π),
        distributions={"0": ConstantProcess(μ=μ_cons), "1": UNormal(σ=σ, μ=μ_norm)},
    )
    disMix = distr.discretize()
    expval = np.array(
        [
            f(disMix.inode(0, j)) * disMix.iweight(0, j)
            for j in range(disMix.n_inodes(0))
        ]
    ).sum()
    """
    The following code is general enough to test different combinations of distributions in a Mixture
    w/out changing many things.
    TO DO: loop over different combinations?
    """
    distr_MC = [
        {"type": "ConstantProcess", "value": μ_cons},
        {"type": np.random.normal, "kwargs": {"loc": μ_norm, "scale": σ}},
    ]
    coef_MC = distr.index.discretize().weights
    M = 1000
    N = len(distr_MC)
    data_MC = np.zeros((M, N))
    for i, d in enumerate(distr_MC):
        if d["type"] == "ConstantProcess":
            data_MC[:, i] = d["value"]
        else:
            data_MC[:, i] = d["type"](size=(M,), **d["kwargs"])
    rand_idx_MC = np.random.choice(np.arange(N), size=(M,), p=coef_MC)
    expval_MC = np.array([f(x) for x in data_MC[np.arange(M), rand_idx_MC]]).sum() / M
    assert abs(expval - expval_MC) < 0.1


def test_UNormal():
    σ = 0.1
    μ = 0.2
    N = 10
    distNorm = UNormal(mu=μ, sigma=σ)
    disNorm_gh = distNorm.discretize()
    disNorm_ep = distNorm.discretize(N=N, method="equiprobable")
    expval_gh = np.array(
        [
            f(disNorm_gh.inode(0, j)) * disNorm_gh.iweight(0, j)
            for j in range(disNorm_gh.n_inodes(0))
        ]
    ).sum()
    expval_ep = np.array(
        [
            f(disNorm_ep.inode(0, j)) * disNorm_ep.iweight(0, j)
            for j in range(disNorm_ep.n_inodes(0))
        ]
    ).sum()
    expval_normal = quad(
        lambda x: f(x)
        / np.sqrt(2 * np.pi * σ ** 2)
        * np.exp(-((x - μ) ** 2) / (2 * σ ** 2)),
        -np.inf,
        np.inf,
    )[0]
    M = 1000
    s_MC = np.random.normal(μ, σ, M)
    expval_MC = np.array([f(s_MC[j]) for j in range(0, M)]).sum() / M
    assert abs(expval_gh - expval_normal) < 0.1
    assert abs(expval_ep - expval_normal) < 0.1


def test_Uniform():
    a = -1
    b = 1
    distUni = Uniform(a, b)
    disUni = distUni.discretize(N=10)
    expval_ep = np.array(
        [
            f(disUni.inode(0, j)) * disUni.iweight(0, j)
            for j in range(disUni.n_inodes(0))
        ]
    ).sum()
    M = 1000
    s_MC = np.random.uniform(a, b, M)
    expval_MC = np.array([f(s_MC[j]) for j in range(0, M)]).sum() / M
    assert abs(expval_ep - expval_MC) < 0.1


def test_Lognormal():
    σ = 0.1
    μ = 0.3
    distLog = LogNormal(μ=μ, σ=σ)
    disLog = distLog.discretize(N=10)
    expval_ep = np.array(
        [
            f(disLog.inode(0, j)) * disLog.iweight(0, j)
            for j in range(disLog.n_inodes(0))
        ]
    ).sum()
    M = 1000
    s_MC = np.random.lognormal(μ, σ, M)
    expval_MC = np.array([f(s_MC[j]) for j in range(0, M)]).sum() / M
    assert abs(expval_ep - expval_MC) < 0.1


def test_beta():
    α = 2
    β = 5
    distbeta = Beta(α, β)
    disbeta = distbeta.discretize(N=10)
    expval_ep = np.array(
        [
            f(disbeta.inode(0, j)) * disbeta.iweight(0, j)
            for j in range(disbeta.n_inodes(0))
        ]
    ).sum()
    M = 1000
    s_MC = np.random.beta(α, β, M)
    expval_MC = np.array([f(s_MC[j]) for j in range(0, M)]).sum() / M
    assert abs(expval_ep - expval_MC) < 0.1



from numpy import atleast_2d, dot
from dolo.numeric.decision_rule import CallableDecisionRule


class TaylorExpansion(CallableDecisionRule):
    def __init__(self, *l):

        if len(l) == 1:
            # backward compatibility
            l = l[0]

        l = tuple(l)

        self.order = len(l) - 2
        self.coefs = l
        self.S_bar = l[0]
        self.X_bar = l[1]
        self.X_s = l[2]
        if self.order >= 2:
            self.X_ss = l[3]
        if self.order >= 3:
            self.X_sss = l[4]
        if self.order >= 4:
            raise Exception("Not implemented")

    def __getitem__(self, ind):

        l = [self.S_bar.copy()]
        l.append(self.X_bar[ind].copy())
        l.append(self.X_s[ind, ...].copy())
        if self.order >= 2:
            l.append(self.X_ss[ind, ...].copy())
        if self.order >= 3:
            l.append(self.X_sss[ind, ...].copy())
        return TaylorExpansion(l)

    def eval_s(self, points):

        if self.order == 1:
            return eval_te_order_1(self.S_bar, self.X_bar, self.X_s, points)
        elif self.order == 2:
            return eval_te_order_2(self.S_bar, self.X_bar, self.X_s, self.X_ss, points)
        elif self.order == 3:
            return eval_te_order_3(
                self.S_bar, self.X_bar, self.X_s, self.X_ss, self.X_sss, points
            )

    def eval_ms(self, m, s):

        return self.eval_s(s)


# helper functions

from numba import guvectorize


@guvectorize(["void(f8[:],f8[:],f8[:,:],f8[:],f8[:])"], "(s),(x),(x,s),(s)->(x)")
def eval_te_order_1(s0, x0, x1, points, out):

    S = points.shape[0]
    X = x0.shape[0]
    for n in range(X):
        out[n] = x0[n]
        for i in range(S):
            out[n] += x1[n, i] * (points[i] - s0[i])

    # return out


@guvectorize(
    ["void(f8[:],f8[:],f8[:,:],f8[:,:,:],f8[:],f8[:])"],
    "(s),(x),(x,s),(x,s,s),(s)->(x)",
)
def eval_te_order_2(s0, x0, x1, x2, points, out):

    S = points.shape[0]
    X = x0.shape[0]
    for n in range(X):
        out[n] = x0[n]
        for i in range(S):
            out[n] += x1[n, i] * (points[i] - s0[i])
            for j in range(S):
                out[n] += x2[n, i, j] * (points[i] - s0[i]) * (points[j] - s0[j]) / 2.0


@guvectorize(
    ["void(f8[:],f8[:],f8[:,:],f8[:,:,:], f8[:,:,:,:], f8[:],f8[:])"],
    "(s),(x),(x,s),(x,s,s),(x,s,s,s),(s)->(x)",
)
def eval_te_order_3(s0, x0, x1, x2, x3, points, out):

    S = points.shape[0]
    X = x0.shape[0]
    for n in range(X):
        out[n] = x0[n]
        for i in range(S):
            out[n] += x1[n, i] * (points[i] - s0[i])
            for j in range(S):
                out[n] += x2[n, i, j] * (points[i] - s0[i]) * (points[j] - s0[j]) / 2.0
                for k in range(S):
                    out[n] += (
                        x3[n, i, j, k]
                        * (points[i] - s0[i])
                        * (points[j] - s0[j])
                        * (points[k] - s0[k])
                        / 6.0
                    )


class CDR(TaylorExpansion):
    pass


def test_taylor_expansion():

    import numpy
    from numpy import array

    s0 = array([0.2, 0.4, 1.1])
    x0 = array([1.2, 0.9])

    N = 1000
    points = numpy.random.random((N, 3))

    X_s = numpy.random.random((2, 3))
    X_ss = numpy.random.random((2, 3, 3))
    X_sss = numpy.random.random((2, 3, 3, 3))
    dr1 = TaylorExpansion(s0, x0, X_s)  # , X_ss, X_sss])
    dr2 = TaylorExpansion(s0, x0, X_s, X_ss)  # , X_sss])
    dr3 = TaylorExpansion(s0, x0, X_s, X_ss, X_sss)

    out1 = dr1(points)
    out2 = dr2(points)
    out3 = dr3(points)

    out1_1d = dr1(points[0, :])
    out2_1d = dr2(points[0, :])
    out3_1d = dr3(points[0, :])

    assert abs(out1_1d - out1[0, :]).max() == 0
    assert abs(out2_1d - out2[0, :]).max() == 0
    assert abs(out3_1d - out3[0, :]).max() == 0

    ds = points - s0[None, :]
    ds2 = ds[:, :, None].repeat(3, axis=2)
    ds3 = ds2[:, :, :, None].repeat(3, axis=3)

    from numpy import dot

    verif1 = x0[None, :] + numpy.dot(ds, X_s.T)
    verif2 = dr2.__call2__(points)
    verif3 = dr3.__call2__(points)

    assert abs(out1 - verif1).max() < 1e-12
    assert abs(out2 - verif2).max() < 1e-12
    assert abs(out3 - verif3).max() < 1e-12


if __name__ == "__main__":

    test_taylor_expansion()



from __future__ import division, print_function, absolute_import

import numpy
from numpy import real_if_close, where
from scipy.linalg import ordqz


def qzordered(A, B, crit=1.0):
    "Eigenvalues bigger than crit are sorted in the top-left."

    TOL = 1e-10

    def select(alpha, beta):
        return alpha ** 2 > crit * beta ** 2

    [S, T, alpha, beta, U, V] = ordqz(A, B, output="real", sort=select)

    eigval = abs(numpy.diag(S) / numpy.diag(T))

    return [S, T, U, V, eigval]


def test_qzordered():

    import numpy

    N = 202
    A = numpy.random.random((N, N))
    B = numpy.random.random((N, N))
    [S, T, U, V, eigval] = qzordered(A, B, 100)


if __name__ == "__main__":
    test_qzordered()



# import ast
# import json
# import ruamel.yaml as ry
# from ruamel.yaml.comments import CommentedSeq
# from dolo.compiler.symbolic import check_expression
# from dolo.compiler.recipes import recipes
# from dolo.misc.termcolor import colored


# class Compare:
#     def __init__(self):
#         self.d = {}

#     def compare(self, A, B):
#         if isinstance(A, ast.Name) and (A.id[0] == '_'):
#             if A.id not in self.d:
#                 self.d[A.id] = B
#                 return True
#             else:
#                 return self.compare(self.d[A.id], B)
#         if not (A.__class__ == B.__class__):
#             return False
#         if isinstance(A, ast.Name):
#             return A.id == B.id
#         elif isinstance(A, ast.Call):
#             if not self.compare(A.func, B.func):
#                 return False
#             if not len(A.args) == len(B.args):
#                 return False
#             for i in range(len(A.args)):
#                 if not self.compare(A.args[i], B.args[i]):
#                     return False
#             return True
#         elif isinstance(A, ast.Num):
#             return A.n == B.n
#         elif isinstance(A, ast.Expr):
#             return self.compare(A.value, B.value)
#         elif isinstance(A, ast.Module):
#             if not len(A.body) == len(B.body):
#                 return False
#             for i in range(len(A.body)):
#                 if not self.compare(A.body[i], B.body[i]):
#                     return False
#             return True
#         elif isinstance(A, ast.BinOp):
#             if not isinstance(A.op, B.op.__class__):
#                 return False
#             if not self.compare(A.left, B.left):
#                 return False
#             if not self.compare(A.right, B.right):
#                 return False
#             return True
#         elif isinstance(A, ast.UnaryOp):
#             if not isinstance(A.op, B.op.__class__):
#                 return False
#             return self.compare(A.operand, B.operand)
#         elif isinstance(A, ast.Subscript):
#             if not self.compare(A.value, B.value):
#                 return False
#             return self.compare(A.slice, B.slice)
#         elif isinstance(A, ast.Index):
#             return self.compare(A.value, B.value)
#         elif isinstance(A, ast.Compare):
#             if not self.compare(A.left, B.left):
#                 return False
#             if not len(A.ops) == len(B.ops):
#                 return False
#             for i in range(len(A.ops)):
#                 if not self.compare(A.ops[i], B.ops[i]):
#                     return False
#             if not len(A.comparators) == len(B.comparators):
#                 return False
#             for i in range(len(A.comparators)):
#                 if not self.compare(A.comparators[i], B.comparators[i]):
#                     return False
#             return True
#         elif isinstance(A, ast.In):
#             return True
#         elif isinstance(A, (ast.Eq, ast.LtE)):
#             return True
#         else:
#             print(A.__class__)
#             raise Exception("Not implemented")


# def compare_strings(a, b):
#     t1 = ast.parse(a)
#     t2 = ast.parse(b)
#     comp = Compare()
#     val = comp.compare(t1, t2)
#     return val


# def match(m, s):
#     if isinstance(m, str):
#         m = ast.parse(m).body[0].value
#     if isinstance(s, str):
#         s = ast.parse(s).body[0].value
#     comp = Compare()
#     val = comp.compare(m, s)
#     d = comp.d
#     if len(d) == 0:
#         return val
#     else:
#         return d


# known_symbol_types = {
#     'dtcc': recipes['dtcc']['symbols'],
# }


# class ModelException(Exception):
#     type = 'error'


# def check_symbol_validity(s):
#     import ast
#     val = ast.parse(s).body[0].value
#     assert (isinstance(val, ast.Name))


# def check_symbols(data):

#     # can raise three types of exceptions
#     # - unknown symbol
#     # - invalid symbol
#     # - already declared

#     # add: not declared if missing 'states', 'controls' ?

#     exceptions = []

#     symbols = data['symbols']
#     cm_symbols = symbols
#     model_type = 'dtcc'

#     already_declared = {}  # symbol: symbol_type, position

#     for key, values in cm_symbols.items():
#         # (start_line, start_column, end_line, end_column) of the key
#         if key not in known_symbol_types[model_type]:
#             l0, c0, l1, c1 = cm_symbols.lc.data[key]
#             exc = ModelException(
#                 "Unknown symbol type '{}'".format(
#                     key, model_type))
#             exc.pos = (l0, c0, l1, c1)
#             # print(l0,c0,l1,c1)
#             exceptions.append(exc)
#             assert (isinstance(values, CommentedSeq))

#         for i, v in enumerate(values):
#             (l0, c0) = values.lc.data[i]
#             length = len(v)
#             l1 = l0
#             c1 = c0 + length
#             try:
#                 check_symbol_validity(v)
#             except:
#                 exc = ModelException("Invalid symbol '{}'".format(v))
#                 exc.pos = (l0, c0, l1, c1)
#                 exceptions.append(exc)
#             if v in already_declared:
#                 ll = already_declared[v]
#                 exc = ModelException(
#                     "Symbol '{}' already declared as '{}'. (pos {})".format(
#                         v, ll[0], (ll[1][0] + 1, ll[1][1])))
#                 exc.pos = (l0, c0, l1, c1)
#                 exceptions.append(exc)
#             else:
#                 already_declared[v] = (key, (l0, c0))

#     return exceptions


# def check_equations(data):

#     model_type = data['model_type']
#     pos0 = data.lc.data['equations']
#     equations = data['equations']

#     exceptions = []
#     recipe = recipes[model_type]
#     specs = recipe['specs']

#     for eq_type in specs.keys():
#         if (eq_type not in equations) and (not specs[eq_type].get(
#                 'optional', True)):
#             exc = ModelException("Missing equation type {}.".format(eq_type))
#             exc.pos = pos0
#             exceptions.append(exc)

#     already_declared = {}
#     unknown = []

#     for eq_type in equations.keys():
#         pos = equations.lc.data[eq_type]
#         if eq_type not in specs:
#             exc = ModelException("Unknown equation type {}.".format(eq_type))
#             exc.pos = pos
#             exceptions.append(exc)
#             unknown.append(eq_type)

#         # BUG: doesn't produce an error when a block is declared twice
#         # should be raised by ruaml.yaml ?
#         elif eq_type in already_declared.keys():
#             exc = ModelException(
#                 "Equation type {} declared twice at ({})".format(eq_type, pos))
#             exc.pos = pos
#             exceptions.append(exc)
#         else:
#             already_declared[eq_type] = pos

#     for eq_type in [k for k in equations.keys() if k not in unknown]:

#         for n, eq in enumerate(equations[eq_type]):
#             eq = eq.replace('<=', '<').replace('==',
#                                                '=').replace('=', '==').replace(
#                                                    '<', '<=')
#             # print(eq)
#             pos = equations[eq_type].lc.data[n]
#             try:
#                 ast.parse(eq)

#             except SyntaxError as e:
#                 exc = ModelException("Syntax Error.")
#                 exc.pos = [
#                     pos[0], pos[1] + e.offset, pos[0], pos[1] + e.offset
#                 ]
#                 exceptions.append(exc)

#         # TEMP: incorrect ordering
#         if specs[eq_type].get('target'):
#             for n, eq in enumerate(equations[eq_type]):
#                 eq = eq.replace('<=', '<').replace('==', '=').replace(
#                     '=', '==').replace('<', '<=')
#                 pos = equations[eq_type].lc.data[n]
#                 lhs_name = str.split(eq, '=')[0].strip()
#                 target = specs[eq_type]['target'][0]
#                 if lhs_name not in data['symbols'][target]:
#                     exc = ModelException(
#                         "Undeclared assignement target '{}'. Add it to '{}'.".
#                         format(lhs_name, target))
#                     exc.pos = [pos[0], pos[1], pos[0], pos[1] + len(lhs_name)]
#                     exceptions.append(exc)
#                 # if n>len(data['symbols'][target]):
#                 else:
#                     right_name = data['symbols'][target][n]
#                     if lhs_name != right_name:
#                         exc = ModelException(
#                             "Left hand side should be '{}' instead of '{}'.".
#                             format(right_name, lhs_name))
#                         exc.pos = [
#                             pos[0], pos[1], pos[0], pos[1] + len(lhs_name)
#                         ]
#                         exceptions.append(exc)
#         # temp
#     return exceptions


# def check_definitions(data):

#     if 'definitions' not in data:
#         return []
#     definitions = data['definitions']
#     if definitions is None:
#         return []

#     exceptions = []
#     known_symbols = sum([[*v] for v in data['symbols'].values()], [])

#     allowed_symbols = {v: (0, ) for v in known_symbols}  # TEMP
#     for p in data['symbols']['parameters']:
#         allowed_symbols[p] = (0, )

#     new_definitions = dict()
#     for k, v in definitions.items():
#         pos = definitions.lc.data[k]
#         if k in known_symbols:
#             exc = ModelException(
#                 'Symbol {} has already been defined as a model symbol.'.format(
#                     k))
#             exc.pos = pos
#             exceptions.append(exc)
#             continue
#         if k in new_definitions:
#             exc = ModelException(
#                 'Symbol {} cannot be defined twice.'.format(k))
#             exc.pos = pos
#             exceptions.append(exc)
#             continue
#         try:
#             check_symbol_validity(k)
#         except:
#             exc = ModelException("Invalid symbol '{}'".format(k))
#             exc.pos = pos
#             exceptions.append(exc)

#             # pos = equations[eq_type].lc.data[n]
#         try:
#             expr = ast.parse(str(v))

#             # print(allowed_symbols)
#             check = check_expression(expr, allowed_symbols)
#             # print(check['problems'])
#             for pb in check['problems']:
#                 name, t, offset, err_type = [pb[0], pb[1], pb[2], pb[3]]
#                 if err_type == 'timing_error':
#                     exc = Exception(
#                         'Timing for variable {} could not be determined.'.
#                         format(pb[0]))
#                 elif err_type == 'incorrect_timing':
#                     exc = Exception(
#                         'Variable {} cannot have time {}. (Allowed: {})'.
#                         format(name, t, pb[4]))
#                 elif err_type == 'unknown_function':
#                     exc = Exception(
#                         'Unknown variable/function {}.'.format(name))
#                 elif err_type == 'unknown_variable':
#                     exc = Exception(
#                         'Unknown variable/parameter {}.'.format(name))
#                 else:
#                     print(err_type)
#                 exc.pos = (pos[0], pos[1] + offset, pos[0],
#                            pos[1] + offset + len(name))
#                 exc.type = 'error'
#                 exceptions.append(exc)

#             new_definitions[k] = v

#             allowed_symbols[k] = (0, )  # TEMP
#             # allowed_symbols[k] = None

#         except SyntaxError as e:
#             pp = pos  # TODO: find right mark for pp
#             exc = ModelException("Syntax Error.")
#             exc.pos = [pp[0], pp[1] + e.offset, pp[0], pp[1] + e.offset]
#             exceptions.append(exc)

#     return exceptions


# def check_calibration(data):
#     # what happens here if symbols are not clean ?
#     symbols = data['symbols']
#     pos0 = data.lc.data['calibration']
#     calibration = data['calibration']
#     exceptions = []
#     all_symbols = []
#     for v in symbols.values():
#         all_symbols += v
#     for s in all_symbols:
#         if (s not in calibration.keys()) and (s not in symbols["exogenous"]):
#             # should skip invalid symbols there
#             exc = ModelException(
#                 "Symbol {} has no calibrated value.".format(s))
#             exc.pos = pos0
#             exc.type = 'warning'
#             exceptions.append(exc)
#     for s in calibration.keys():
#         val = str(calibration[s])
#         try:
#             ast.parse(val)
#         except SyntaxError as e:
#             pos = calibration.lc.data[s]
#             exc = ModelException("Syntax Error.")
#             exc.pos = [pos[0], pos[1] + e.offset, pos[0], pos[1] + e.offset]
#             exceptions.append(exc)
#     return exceptions


# def check_all(data):
#     def serious(exsc):
#         return ('error' in [e.type for e in exsc])

#     exceptions = check_infos(data)
#     if serious(exceptions):
#         return exceptions
#     exceptions = check_symbols(data)
#     if serious(exceptions):
#         return exceptions
#     exceptions += check_definitions(data)
#     if serious(exceptions):
#         return exceptions
#     exceptions += check_equations(data)
#     if serious(exceptions):
#         return exceptions
#     exceptions += check_calibration(data)
#     if serious(exceptions):
#         return exceptions
#     return exceptions


# def human_format(err):
#     err_type = err['type']
#     err_type = colored(
#         err_type, color=('red' if err_type == 'error' else 'yellow'))
#     err_range = str([e + 1 for e in err['range'][0]])[1:-1]
#     return '{:7}: {:6}: {}'.format(err_type, err_range, err['text'])


# def check_infos(data):
#     exceptions = []
#     if 'model_type' in data:
#         model_type = data['model_type']
#         if model_type not in ['dtcc', 'dtmscc', 'dtcscc', 'dynare']:
#             exc = ModelException('Uknown model type: {}.'.format(
#                 str(model_type)))
#             exc.pos = data.lc.data['model_type']
#             exc.type = 'error'
#             exceptions.append(exc)
#     else:
#         model_type = 'dtcc'
#         data['model_type'] = 'dtcc'
#         # exc = ModelException("Missing field: 'model_type'.")
#         # exc.pos = (0,0,0,0)
#         # exc.type='error'
#         # exceptions.append(exc)
#     if 'name' not in data:
#         exc = ModelException("Missing field: 'name'.")
#         exc.pos = (0, 0, 0, 0)
#         exc.type = 'warning'
#         exceptions.append(exc)
#     return exceptions


# def lint(txt, source='<string>', format='human', catch_exception=False):

#     # raise ModelException if it doesn't work correctly
#     if isinstance(txt, str):
#         try:
#             data = ry.load(txt, ry.RoundTripLoader)
#         except Exception as exc:
#             if not catch_exception:
#                 raise exc
#             return []  # should return parse error
#     else:
#         # txt is then assumed to be a ruamel structure
#         data = txt

#     if not ('symbols' in data or 'equations' in data or 'calibration' in data):
#         # this is probably not a yaml filename
#         output = []
#     else:
#         try:
#             exceptions = check_all(data)
#         except Exception as e:
#             if not catch_exception:
#                 raise(e)
#             exc = ModelException("Linter Error: Uncaught Exception.")
#             exc.pos = [0, 0, 0, 0]
#             exc.type = 'error'
#             exceptions = [exc]

#         output = []
#         for k in exceptions:
#             try:
#                 err_type = k.type
#             except:
#                 err_type = 'error'
#             output.append({
#                 'type':
#                 err_type,
#                 'source':
#                 source,
#                 'range': ((k.pos[0], k.pos[1]), (k.pos[2], k.pos[3])),
#                 'text':
#                 k.args[0]
#             })

#     if format == 'json':
#         return (json.dumps(output))
#     elif format == 'human':
#         return (str.join("\n", [human_format(e) for e in output]))
#     elif not format:
#         return output
#     else:
#         raise ModelException("Unkown format {}.".format(format))


# TODO:
# - check name (already defined by smbdy else ?)
# - description: ?
# - calibration:
#      - incorrect key
#          - warning if not a known symbol ?
#          - not a recognized identifier
#          - defined twice
#      - impossible to solve in closed form (depends on ...)
#      - incorrect equation
#           - grammatically incorrect
#           - contains timed variables
#      - warnings:
#           - missing values
# - equations: symbols already known (beware of speed issues)
#     - unknown group of equations
#     - incorrect syntax
#     - undeclared variable (and not a function)
#     - indexed parameter
#     - incorrect order
#     - incorrect complementarities
#     - incorrect recipe: unexpected symbol type
#     - nonzero residuals (warning, to be done without compiling)
# - options: if present
#     - approximation_space:
#          - inconsistent boundaries
#                - must equal number of states
#     - distribution:
#          - same size as shocks



import numpy as np
from numpy import sin
import scipy as sp
from scipy.signal import fftconvolve
from scipy.sparse import linalg as spla


def hp_filter(data, lam=1600):
    """
    This function will apply a Hodrick-Prescott filter to a dataset.
    The return value is the filtered data-set found according to:
        min sum((X[t] - T[t])**2 + lamb*((T[t+1] - T[t]) - (T[t] - T[t-1]))**2)
          T   t

    T = Lambda**-1 * Y

    Parameters
    ----------
        data: array, dtype=float
            The data set for which you want to apply the HP_filter.
            This mustbe a numpy array.
        lamb: array, dtype=float
            This is the value for lambda as used in the equation.

    Returns
    -------
        T: array, dtype=float
            The solution to the minimization equation above (the trend).
        Cycle: array, dtype=float
            This is the 'stationary data' found by Y - T.

    Notes
    -----
        This function implements sparse methods to be efficient enough to handle
        very large data sets.
    """

    Y = np.asarray(data)

    if Y.ndim == 2:
        resp = [hp_filter(e) for e in data]
        T = np.row_stack([e[0] for e in resp])
        Cycle = np.row_stack([e[1] for e in resp])
        return [T, Cycle]

    elif Y.ndim > 2:
        raise Exception("HP filter is not defined for dimension >= 3.")

    lil_t = len(Y)
    big_Lambda = sp.sparse.eye(lil_t, lil_t)
    big_Lambda = sp.sparse.lil_matrix(big_Lambda)

    # Use FOC's to build rows by group. The first and last rows are similar.
    # As are the second-second to last. Then all the ones in the middle...
    first_last = np.array([1 + lam, -2 * lam, lam])
    second = np.array([-2 * lam, (1 + 5 * lam), -4 * lam, lam])
    middle_stuff = np.array([lam, -4.0 * lam, 1 + 6 * lam, -4 * lam, lam])

    # --------------------------- Putting it together --------------------------#

    # First two rows
    big_Lambda[0, 0:3] = first_last
    big_Lambda[1, 0:4] = second

    # Last two rows. Second to last first : we have to reverse arrays
    big_Lambda[lil_t - 2, -4:] = second[::-1]
    big_Lambda[lil_t - 1, -3:] = first_last[::-1]

    # Middle rows
    for i in range(2, lil_t - 2):
        big_Lambda[i, i - 2 : i + 3] = middle_stuff

    # spla.spsolve requires csr or csc matrix. I choose csr for fun.
    big_Lambda = sp.sparse.csr_matrix(big_Lambda)

    T = spla.spsolve(big_Lambda, Y)

    Cycle = Y - T

    return T, Cycle


def bandpass_filter(data, k, w1, w2):
    """
    This function will apply a bandpass filter to data. It will be kth
    order and will select the band between w1 and w2.

    Parameters
    ----------
        data: array, dtype=float
            The data you wish to filter
        k: number, int
            The order of approximation for the filter. A max value for
            this isdata.size/2
        w1: number, float
            This is the lower bound for which frequencies will pass
            through.
        w2: number, float
            This is the upper bound for which frequencies will pass
            through.

    Returns
    -------
        y: array, dtype=float
            The filtered data.
    """
    data = np.asarray(data)
    low_w = np.pi * 2 / w2
    high_w = np.pi * 2 / w1
    bweights = np.zeros(2 * k + 1)
    bweights[k] = (high_w - low_w) / np.pi
    j = np.arange(1, int(k) + 1)
    weights = 1 / (np.pi * j) * (sin(high_w * j) - sin(low_w * j))
    bweights[k + j] = weights
    bweights[:k] = weights[::-1]

    bweights -= bweights.mean()

    return fftconvolve(bweights, data, mode="valid")



def get_derivatives(model, steady_state=None):
    """The sum of two numbers."""

    from dolo.numeric.processes import VAR1, IIDProcess
    from dolo.numeric.distribution import Normal

    import numpy as np

    if steady_state is None:
        steady_state = model.calibration

    m = steady_state["exogenous"]
    s = steady_state["states"]
    x = steady_state["controls"]
    p = steady_state["parameters"]

    f = model.functions["arbitrage"]
    g = model.functions["transition"]

    n_m = len(m)
    n_s = len(s)
    n_x = len(x)

    f_, f_m, f_s, f_x, f_M, f_S, f_X = f(m, s, x, m, s, x, p, diff=True)
    g_, g_m, g_s, g_x, g_M = g(m, s, x, m, p, diff=True)

    process = model.exogenous

    if isinstance(process, VAR1):
        A = process.rho
        B = np.eye(A.shape[0])
        G_s = np.row_stack(
            [
                np.column_stack([A, np.zeros((n_m, n_s))]),
                np.column_stack([g_m + g_M @ A, g_s]),
            ]
        )
        G_x = np.row_stack([np.zeros((n_m, n_x)), g_x])
        G_e = g_m @ B
        F_s = np.column_stack([f_m, f_s])
        F_S = np.column_stack([f_M, f_S])
        F_x = f_x
        F_X = f_X
    elif isinstance(process, IIDProcess):
        G_s = g_s
        G_x = g_x
        G_e = g_m
        F_s = f_s
        F_S = f_S
        F_x = f_x
        F_X = f_X
    else:
        raise Exception(f"Not implemented: perturbation for shock {process.__class__}")

    return G_s, G_x, G_e, F_s, F_x, F_S, F_X


from dolo.numeric.grids import PointGrid, EmptyGrid
from dolo.numeric.decision_rule import CallableDecisionRule


class BivariateTaylor(CallableDecisionRule):
    def __init__(self, m_bar, s_bar, x_bar, C_m, C_s):

        self.endo_grid = PointGrid(s_bar)
        if C_m is None:
            self.exo_grid = EmptyGrid()
        else:
            self.exo_grid = PointGrid(m_bar)

        self.m_bar = m_bar
        self.s_bar = s_bar
        self.x_bar = x_bar
        self.C_m = C_m
        self.C_s = C_s

    def eval_s(self, s):

        s = np.array(s)
        if s.ndim == 1:
            return self.eval_s(s[None, :])[0, :]
        return self.eval_ms(None, s)

    def eval_ms(self, m, s):

        m = np.array(m)
        s = np.array(s)
        if m.ndim == 1 and s.ndim == 1:
            if self.C_m is not None:
                return self.eval_ms(m[None, :], s[None, :])[0, :]
            else:
                return self.eval_ms(None, s[None, :])[0, :]
        elif m.ndim == 1:
            m = m[None, :]
        elif s.ndim == 1:
            s = s[None, :]

        C_m = self.C_m
        C_s = self.C_s
        if C_m is not None:
            dm = m - self.m_bar[None, :]
            ds = s - self.s_bar[None, :]
            return self.x_bar[None, :] + dm @ C_m.T + ds @ C_s.T
        else:
            ds = s - self.s_bar[None, :]
            return self.x_bar[None, :] + ds @ C_s.T


import numpy as np
from numpy import column_stack, dot, eye, row_stack, zeros
from numpy.linalg import solve

from dolo.numeric.extern.qz import qzordered


def approximate_1st_order(g_s, g_x, g_e, f_s, f_x, f_S, f_X):

    n_s = g_s.shape[0]  # number of controls
    n_x = g_x.shape[1]  # number of states
    n_e = g_e.shape[1]
    n_v = n_s + n_x

    A = row_stack(
        [column_stack([eye(n_s), zeros((n_s, n_x))]), column_stack([-f_S, -f_X])]
    )

    B = row_stack([column_stack([g_s, g_x]), column_stack([f_s, f_x])])

    [S, T, Q, Z, eigval] = qzordered(A, B, 1.0 - 1e-8)

    Z = Z.real

    diag_S = np.diag(S)
    diag_T = np.diag(T)

    tol_geneigvals = 1e-10

    try:
        ok = sum((abs(diag_S) < tol_geneigvals) * (abs(diag_T) < tol_geneigvals)) == 0
        assert ok
    except Exception as e:
        raise GeneralizedEigenvaluesError(diag_S=diag_S, diag_T=diag_T)

    eigval_s = sorted(eigval, reverse=False)
    if max(eigval[:n_s]) >= 1 and min(eigval[n_s:]) < 1:
        # BK conditions are met
        pass
    else:
        ev_a = eigval_s[n_s - 1]
        ev_b = eigval_s[n_s]
        cutoff = (ev_a - ev_b) / 2
        if not ev_a > ev_b:
            raise GeneralizedEigenvaluesSelectionError(
                A=A,
                B=B,
                eigval=eigval,
                cutoff=cutoff,
                diag_S=diag_S,
                diag_T=diag_T,
                n_states=n_s,
            )
        import warnings

        if cutoff > 1:
            warnings.warn("Solution is not convergent.")
        else:
            warnings.warn(
                "There are multiple convergent solutions. The one with the smaller eigenvalues was selected."
            )
        [S, T, Q, Z, eigval] = qzordered(A, B, cutoff)

    Z11 = Z[:n_s, :n_s]
    # Z12 = Z[:n_s, n_s:]
    Z21 = Z[n_s:, :n_s]
    # Z22 = Z[n_s:, n_s:]
    # S11 = S[:n_s, :n_s]
    # T11 = T[:n_s, :n_s]

    # first order solution
    # P = (solve(S11.T, Z11.T).T @ solve(Z11.T, T11.T).T)
    C = solve(Z11.T, Z21.T).T

    A = g_s + g_x @ C
    B = g_e

    return C, eigval_s


from .results import AlgoResult, PerturbationResult


def perturb(
    model,
    verbose=False,
    steady_state=None,
    eigmax=1.0 - 1e-6,
    solve_steady_state=False,
    order=1,
    details=True,
):
    """Compute first order approximation of optimal controls

    Parameters:
    -----------

    model: NumericModel
        Model to be solved

    verbose: boolean
        If True: displays number of contracting eigenvalues

    steady_state: ndarray
        Use supplied steady-state value to compute the approximation.
        The routine doesn't check whether it is really a solution or not.

    solve_steady_state: boolean
        Use nonlinear solver to find the steady-state

    orders: {1}
        Approximation order. (Currently, only first order is supported).

    Returns:
    --------

    TaylorExpansion:
        Decision Rule for the optimal controls around the steady-state.

    """

    if order > 1:
        raise Exception("Not implemented.")

    if steady_state is None:
        steady_state = model.calibration

    G_s, G_x, G_e, F_s, F_x, F_S, F_X = get_derivatives(
        model, steady_state=steady_state
    )

    C, eigvals = approximate_1st_order(G_s, G_x, G_e, F_s, F_x, F_S, F_X)

    m = steady_state["exogenous"]
    s = steady_state["states"]
    x = steady_state["controls"]

    from dolo.numeric.processes import VAR1, IIDProcess
    from dolo.numeric.distribution import MvNormal

    process = model.exogenous

    if isinstance(process, VAR1):
        C_m = C[:, : len(m)]
        C_s = C[:, len(m) :]
    elif isinstance(process, IIDProcess):
        C_m = None
        C_s = C
    dr = BivariateTaylor(m, s, x, C_m, C_s)
    if not details:
        return dr
    else:
        return PerturbationResult(
            dr,
            eigvals,
            True,  # otherwise an Exception should have been raised already
            True,  # otherwise an Exception should have been raised already
            True,  # otherwise an Exception should have been raised already
        )



import unittest


class Counter:
    n = 0


from dolo.misc.caching import clear_cache, cachedondisk, memoized, hashable

clear_cache()


@cachedondisk
def fun(a, b):
    Counter.n += 1
    return a + b


import dolo.config


class CachingTestCase(unittest.TestCase):
    def test_cacheondisk(self):

        Counter.n = 0
        res1 = fun(1, 2)
        assert Counter.n == 1
        res2 = fun(1, 2)
        assert Counter.n == 1

    def test_cache_arrays(self):

        import numpy

        A = numpy.array([1, 2, 3, 4]).reshape((2, 2))
        B = numpy.array([0.1, 0.2, 0.3, 0.4]).reshape((2, 2))
        Counter.n = 0
        res1 = fun(A, B)
        assert Counter.n == 1
        res2 = fun(A, B)
        assert Counter.n == 1


if __name__ == "__main__":
    unittest.main()



import functools

from dolo.misc.caching import memoized, cachedondisk

import sys

is_python_3 = sys.version_info >= (3, 0)


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    import warnings

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        if is_python_3:
            code = func.__code__
        else:
            code = func.func_code
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=Warning,
            filename=code.co_filename,
            lineno=code.co_firstlineno + 1,
        )
        return func(*args, **kwargs)

    return new_func


@deprecated
def test_deprecation():
    pass



import unittest

import numpy as np

from dolo.numeric.optimize.ncpsolve import ncpsolve


def josephy(x):
    #   Computes the function value F(x) of the NCP-example by Josephy.
    n = len(x)
    Fx = np.zeros(n)
    Fx[0] = 3 * x[0] ** 2 + 2 * x[0] * x[1] + 2 * x[1] ** 2 + x[2] + 3 * x[3] - 6
    Fx[1] = 2 * x[0] ** 2 + x[0] + x[1] ** 2 + 3 * x[2] + 2 * x[3] - 2
    Fx[2] = 3 * x[0] ** 2 + x[0] * x[1] + 2 * x[1] ** 2 + 2 * x[2] + 3 * x[3] - 1
    Fx[3] = x[0] ** 2 + 3 * x[1] ** 2 + 2 * x[2] + 3 * x[3] - 3
    return Fx


def Djosephy(x):
    # Local Variables: x, DFx, n
    # Function calls: Djosephy, zeros, length
    #%
    #%   Computes the Jacobian DF(x) of the NCP-example by Josephy
    #%
    n = len(x)
    DFx = np.zeros((n, n))
    DFx[0, 0] = 6.0 * x[0] + 2.0 * x[1]
    DFx[0, 1] = 2.0 * x[0] + 4.0 * x[1]
    DFx[0, 2] = 1.0
    DFx[0, 3] = 3.0
    DFx[1, 0] = 4.0 * x[0] + 1.0
    DFx[1, 1] = 2.0 * x[1]
    DFx[1, 2] = 3.0
    DFx[1, 3] = 2.0
    DFx[2, 0] = 6.0 * x[0] + x[1]
    DFx[2, 1] = x[0] + 4.0 * x[1]
    DFx[2, 2] = 2.0
    DFx[2, 3] = 3.0
    DFx[3, 0] = 2.0 * x[0]
    DFx[3, 1] = 6.0 * x[1]
    DFx[3, 2] = 2.0
    DFx[3, 3] = 3.0
    return DFx


class SerialSolve(unittest.TestCase):
    #
    # def test_serial_smooth(self):
    #
    #     x0 = np.array([0.5,0.5,0.5,0.5])
    #
    #
    #     lb = np.array([0.0,0.6,0.0,0.0])
    #     ub = np.array([1.0,1.0,1.0,0.4])
    #
    #     fval = np.array([ 0.5, 0.5, 0.1,0.5 ])
    #
    #     jac = np.array([
    #         [1.0,0.2,0.1,0.0],
    #         [1.0,0.2,0.1,0.0],
    #         [0.0,1.0,0.2,0.0],
    #         [0.1,1.0,0.2,0.1]
    #     ])
    #
    #     N = 10
    #     d = len(fval)
    #
    #     s_x0 = np.row_stack([x0]*N)
    #     s_lb = np.row_stack([lb]*N)
    #     s_ub = np.row_stack([ub]*N)
    #     s_fval = np.row_stack([fval]*N)
    #
    #
    #     s_jac = np.zeros( (N,d,d) )
    #     for i in range(N):
    #         s_jac[i,:,:] = jac
    #
    #     from dolo.numeric.ncpsolve import smooth
    #
    #     [fnew, Jnew] = smooth(x0, lb, ub, fval, J=jac)
    #
    #     serial_fnew_true = np.zeros( (N,d) )
    #     serial_Jnew_true = np.zeros( (N,d,d) )
    #     for n in range(N):
    #         serial_fnew_true[n,:] = fnew
    #         serial_Jnew_true[n,:,:] = Jnew
    #
    #
    #     [serial_fnew, serial_Jnew] = smooth(s_x0, s_lb, s_ub, s_fval, J=s_jac, serial=True)
    #
    #     from numpy.testing import assert_equal
    #
    #     assert_equal(serial_fnew, serial_fnew_true)
    #     assert_equal(serial_Jnew, serial_Jnew_true)

    def test_serial_solve(self):

        fun = lambda x: [-josephy(x), -Djosephy(x)]

        x0 = np.array([1.25, 0.01, 0.01, 0.50])
        lb = np.array([0.00, 0.00, 0.00, 0.00])
        # ub=np.array( [inf, inf, inf, inf] )
        ub = np.array([1e20, 1e20, 1e20, 1e20])

        # resp = ncpsolve(fun,  lb, ub, x0, tol=1e-15)

        sol = np.array([1.22474487e00, 0.00000000e00, 3.60543164e-17, 5.00000000e-01])

        # assert_almost_equal(sol,  resp)

        N = 10
        d = len(x0)

        s_x0 = np.row_stack([x0] * N)
        s_lb = np.row_stack([lb] * N)
        s_ub = np.row_stack([ub] * N)

        def serial_fun(xvec):

            resp = np.zeros((N, d))
            dresp = np.zeros((N, d, d))
            for n in range(N):
                [v, dv] = fun(xvec[n, :])
                resp[n, :] = v
                dresp[n, :, :] = dv
            return [resp, dresp]

        res = serial_fun(s_x0)[0]

        serial_sol = ncpsolve(
            serial_fun, s_lb, s_ub, s_x0, jactype="serial", verbose=True
        )

        print(serial_sol)


if __name__ == "__main__":
    from numpy import inf

    fun = lambda x: [-josephy(x), -Djosephy(x)]

    x0 = np.array([1.25, 0.01, 0.01, 0.50])
    lb = np.array([0.00, 0.00, 0.00, 0.00])
    ub = np.array([inf, inf, inf, inf])

    resp = ncpsolve(fun, lb, ub, x0, tol=1e-15)

    sol = np.array([1.22474487e00, 0.00000000e00, 3.60543164e-17, 5.00000000e-01])

    from numpy.testing import assert_almost_equal

    assert_almost_equal(sol, resp[0])

    N = 2
    d = len(x0)

    serial_sol_check = np.zeros((d, N))
    for n in range(N):
        serial_sol_check[:, n] = resp[0]

    s_x0 = np.column_stack([x0] * N)
    s_lb = np.column_stack([lb] * N)
    s_ub = np.column_stack([ub] * N)

    def serial_fun(xvec):

        resp = np.zeros((d, N))
        dresp = np.zeros((d, d, N))
        for n in range(N):
            [v, dv] = fun(xvec[:, n])
            resp[:, n] = v
            dresp[:, :, n] = dv
        return [resp, dresp]

    serial_sol = ncpsolve(serial_fun, s_lb, s_ub, s_x0, serial=True)

    err = abs(serial_sol[0] - serial_sol_check).max()

    assert_almost_equal(serial_sol[0], serial_sol_check)



from numba import jit
from numpy.linalg import solve
from numpy import zeros_like
from numba import guvectorize, float64, void, generated_jit
import numpy
from numba import generated_jit


def swaplines_tensor(i, j, M):
    n0, n1, n2 = M.shape
    for k in range(n1):
        for l in range(n2):
            t = M[i, k, l]
            M[i, k, l] = M[j, k, l]
            M[j, k, l] = t


def swaplines_matrix(i, j, M):
    n = M.shape[1]
    for k in range(n):
        t = M[i, k]
        M[i, k] = M[j, k]
        M[j, k] = t


def swaplines_vector(i, j, M):
    n = M.shape[0]
    t = M[i]
    M[i] = M[j]
    M[j] = t


@generated_jit(cache=True)
def swaplines(i, j, M):
    if M.ndim == 1:
        return swaplines_vector
    elif M.ndim == 2:
        return swaplines_matrix
    elif M.ndim == 3:
        return swaplines_tensor


def substract_tensor(i, j, c, M):
    # Li <- Li - c*Lj
    n0, n1, n2 = M.shape
    for k in range(n1):
        for l in range(n2):
            M[i, k, l] = M[i, k, l] - c * M[j, k, l]


def substract_matrix(i, j, c, M):
    # Li <- Li - c*Lj
    n = M.shape[0]
    for k in range(n):
        M[i, k] = M[i, k] - c * M[j, k]


def substract_vector(i, j, c, M):
    # Li <- Li - c*Lj
    # n = M.shape[0]
    M[i] = M[i] - c * M[j]


@generated_jit(cache=True)
def substract(i, j, c, M):
    if M.ndim == 1:
        return substract_vector
    elif M.ndim == 2:
        return substract_matrix
    elif M.ndim == 3:
        return substract_tensor


def divide_tensor(i, c, M):
    # Li <- Li - c*Lj
    n0, n1, n2 = M.shape
    for k in range(n1):
        for l in range(n2):
            M[i, k, l] /= c


def divide_matrix(i, c, M):
    # Li <- Li - c*Lj
    n = M.shape[0]
    for k in range(n):
        M[i, k] /= c


def divide_vector(i, c, M):
    # Li <- Li - c*Lj
    M[i] /= c


@generated_jit(cache=True)
def divide(i, c, M):
    if M.ndim == 1:
        return divide_vector
    elif M.ndim == 2:
        return divide_matrix
    elif M.ndim == 3:
        return divide_tensor


@jit(nopython=True, cache=True)
def invert(A, B):

    # inverts A and puts result in B (modifies inputs)
    n = A.shape[0]
    #
    for i in range(n):
        # print(A,B)
        # find pivot
        max_err = -1.0
        max_i = 0
        for i0 in range(i, n):
            err = abs(A[i0, i])
            if err >= max_err:
                max_err = err
                max_i = i0

        swaplines(i, max_i, A)
        swaplines(i, max_i, B)

        c = A[i, i]
        divide(i, c, A)
        divide(i, c, B)

        for i0 in range(i + 1, n):
            f = A[i0, i]
            substract(i0, i, f, A)
            substract(i0, i, f, B)

    for i in range(n - 1, -1, -1):
        for i0 in range(i):
            f = A[i0, i]
            substract(i0, i, f, A)
            substract(i0, i, f, B)


target = "parallel"


@guvectorize(
    [(float64[:, :], float64[:, :])],
    "(n,n)->(n,n)",
    nopython=True,
    target=target,
    cache=True,
)
def invert_gu(A, Ainv):
    Ainv[:, :] = 0
    n = A.shape[0]
    for i in range(n):
        Ainv[i, i] = 1.0
    invert(A, Ainv)


@guvectorize(
    [(float64[:, :], float64[:])],
    "(n,n)->(n)",
    nopython=True,
    target=target,
    cache=True,
)
def solve_gu(A, V):
    n = A.shape[0]
    invert(A, V)


@guvectorize(
    [(float64[:, :], float64[:, :], float64[:])],
    "(n,n),(n,p)->()",
    nopython=True,
    target=target,
    cache=True,
)
def solve_tensor(A, V, dum):
    n = A.shape[0]
    invert(A, V)


@guvectorize(
    [(float64[:, :], float64[:, :, :], float64[:])],
    "(n,n),(n,p,q)->()",
    nopython=True,
    target=target,
    cache=True,
)
def solve_tensor_old(A, V, dum):
    n = A.shape[0]
    invert(A, V)


def test_list_of_matrices():
    import numpy

    N = 10
    m = 4
    A0 = numpy.random.random((N, m, m))
    A = A0.copy()
    B = invert_gu(A)
    err_max = 0.0
    for i in range(N):
        mat = A0[i, :, :]
        matinv = B[i, :, :]
        err = abs(mat @ matinv - numpy.eye(m)).max()
        err_2 = abs(A - numpy.eye(m)).max()
        assert err < 1e-11
        assert err_2 < 1e-11


@jit
def serial_solve(A, B, diagnose=True):
    sol = zeros_like(B)
    for i in range(sol.shape[0]):
        sol[i, :] = solve(A[i, :, :], B[i, :])
    return sol


@jit(nopython=True)
def mult_AB(dres, jres):
    n_m, N, n_x, n_xx = dres.shape
    assert jres.shape == (n_m, N, n_x, n_m, n_x)
    out = jres.copy()
    for i in range(n_m):
        for j in range(n_m):
            for v in range(n_x):
                out[i, :, :, j, v] = serial_solve(dres[i, :, :, :], jres[i, :, :, j, v])
    return out


def test_list_of_arrays():
    import numpy

    m = 4
    N = 10
    x = 2
    A0 = numpy.random.random((m, N, x, x))
    B0 = numpy.random.random((m, N, x, m, x))
    A = A0.copy()
    B = B0.copy()
    C = solve_tensor(A, B)
    C2 = mult_AB(A0, B0)

    print(abs(C2 - B).max())


#
# test_list_of_arrays()
#
# test_list_of_matrices()


# if False:
#
#
#     A = numpy.random.random((3,3))
#     P = A.copy()
#     Q = numpy.eye(A.shape[0])
#     invert(P,Q)
#
#     A = numpy.random.random((3,3))
#     V = numpy.random.random((3))
#     P = A.copy()
#     W = V.copy()
#     invert(P,W)
#
#     N = 10000
#     m = 9
#     x = 8
#     A0 = numpy.random.random((N,m,8,8))
#     V0 = numpy.random.random((N,m,8))
#     A = A0.copy()
#     V = V0.copy()
#
#
#     from serial_operations import serial_solve
#     A = A0.copy()
#     V = V0.copy()
#
#
#     N = 10000
#     m = 27
#     x = 8
#     A0 = numpy.random.random((N,m,x,x))
#     V0 = numpy.random.random((N,m,x,m,x))
#     A = A0.copy()
#     V = V0.copy()
#
#     i = 5
#     A0[i,:,:]@V[i,:,0,2]
#
#     V0[i,:,0,2]



from itertools import product

import numpy as np


def numdiff1(f, x0, dv=1e-8):
    """Returns the derivative of f w.r.t. to multidimensional vector x0
    If x0 is of dimension R1 x ... x Rd x Rn dimension of f is assumed to be
    in the form S1 x ... x Sf x Rn. The last dimension corresponds to various
    observations. The value returned is of dimension :
    S1 x ... x Sf x R1 x ... x Rd x Rn
    """
    in_shape = x0.shape
    nobs = in_shape[-1]
    dd = in_shape[:-1]
    f0 = f(x0)
    assert f0.shape[-1] == nobs
    f_shape = f0.shape[:-1]

    out_shape = f_shape + dd + (nobs,)
    ret = np.zeros(out_shape)

    for ind in product(*[range(i) for i in dd]):
        sl = ind + (slice(None, None, None),)
        x = x0.copy()
        x[sl] += dv
        x2 = x0.copy()
        x2[sl] -= dv
        df = (f(x) - f(x2)) / dv / 2.0
        obj = [Ellipsis] + list(ind) + [slice(None, None, None)]
        obj = tuple(obj)
        ret[obj] = df

    return ret


def numdiff2(f, x0, dv=1e-8):
    """Returns the derivative of f w.r.t. to multidimensional vector x0
    If x0 is of dimension R1 x ... x Rd dimension of f is assumed to be
    in the form S1 x ... x Sf x Rn. The last dimension corresponds to various
    observations. The value returned is of dimension :
    S1 x ... x Sf x R1 x ... x Rd x Rn
    """

    dd = x0.shape
    f0 = f(x0)
    nobs = f0.shape[-1]
    f_shape = f0.shape[:-1]

    out_shape = f_shape + dd + (nobs,)
    ret = np.zeros(out_shape)

    for ind in product(*[range(i) for i in dd]):
        x = x0.copy()
        x[ind] += dv
        x2 = x0.copy()
        x2[ind] -= dv
        df = (f(x) - f(x2)) / dv / 2.0
        obj = [Ellipsis] + list(ind) + [slice(None, None, None)]
        # obj = tuple(obj)
        ret[obj] = df

    return ret


import numpy
from numba import jit


import numpy
from numba import guvectorize
from numba import double


def solve(m, sol):

    # seems to segfault on windows

    h, w = m.shape

    for y in range(0, h):
        maxrow = y
        for y2 in range(y + 1, h):  # Find max pivot
            if abs(m[y2, y]) > abs(m[maxrow, y]):
                maxrow = y2
        for y2 in range(0, w):
            t = m[y, y2]
            m[y, y2] = m[maxrow, y2]
            m[maxrow, y2] = t

        for y2 in range(y + 1, h):  # Eliminate column y
            c = m[y2, y] / m[y, y]
            for x in range(y, w):
                m[y2, x] -= m[y, x] * c

    for y in range(h - 1, 0 - 1, -1):  # Backsubstitute
        c = m[y, y]
        for y2 in range(0, y):
            for x in range(w - 1, y - 1, -1):
                m[y2, x] -= m[y, x] * m[y2, y] / c
        m[y, y] /= c
        for x in range(h, w):  # Normalize row y
            m[y, x] /= c

    for y in range(h):
        sol[y] = m[y, w - 1]


# serial_solve_numba = guvectorize('void(f8[:,:], f8[:])', '(m,n)->(m)')(solve)
serial_solve_numba = guvectorize(
    "void(f8[:,:], f8[:])", "(m,n)->(m)", target="parallel"
)(solve)


from numpy.linalg import solve as linalg_solve
from numpy import zeros_like


def serial_solve(A, B, diagnose=True):

    if diagnose:

        sol = zeros_like(B)

        for i in range(sol.shape[0]):
            try:
                sol[i, :] = linalg_solve(A[i, :, :], B[i, :])
            except:
                # Should be a special type of exception
                a = Exception("Error solving point {}".format(i))
                a.x = B[i, :]
                a.J = A[i, :, :]
                a.i = i
                raise a

    else:
        M = numpy.concatenate([A, B[:, :, None]], axis=2)
        sol = numpy.zeros_like(B)
        serial_solve_numba(M, sol)

    return sol


@jit(cache=True)
def serial_multiplication(A, B):

    if A.ndim == 2 and B.ndim == 2:
        return numpy.dot(A, B)

    I = A.shape[1]
    J = A.shape[2]
    N = A.shape[0]
    K = B.shape[2]

    assert B.shape[1] == J
    assert B.shape[0] == N

    v = 0.0
    resp = np.zeros((N, I, K))
    for n in range(N):
        for i in range(I):
            for k in range(K):
                v = 0
                for j in range(J):
                    v += A[n, i, j] * B[n, j, k]
                resp[n, i, k] = v

    return resp


# def serial_multiplication_vector(A,X):
#
#     I = A.shape[0]
#     J = A.shape[1]
#     N = A.shape[2]
#
#     assert(X.shape[0]==J)
#     assert(X.shape[1]==N)
#
#     resp = np.zeros( (I,N) )
#     for i in range(I):
# #        T = np.zeros( N )
#         for j in range(J):
# #            T += A[i,j,:]*B[j,k,:]
#             resp[i,:] += A[i,j,:]*X[j,:]
#     return resp


if __name__ == "__main__":

    import numpy.random

    A = numpy.random.random((I, J, N))
    B = numpy.random.random((J, K, N))

    import time

    r = time.time()
    C0 = serial_multiplication(A, B)
    s = time.time()
    t = time.time()
    C2 = smult(A, B)
    u = time.time()

    AA = numpy.rollaxis(A, 2).copy()
    BB = numpy.rollaxis(B, 2).copy()

    x = time.time()
    CC = numpy.zeros((A.shape[2], A.shape[0], B.shape[1]))
    serial_mult_numba(AA, BB, CC)
    y = time.time()

    print("Py : {}".format(s - r))
    print("Cython : {}".format(u - t))
    print("Numba : {}".format(y - x))

    print(abs(C2 - C0).max())



import pandas as pd

import numpy as np


default_layout = {
    1: [(4, 4), (1, 1)],
    2: [(6, 4), (1, 2)],
    3: [(8, 4), (1, 3)],
    4: [(8, 8), (2, 2)],
    5: [(8, 8), (2, 3)],
    6: [(8, 8), (2, 3)],
    7: [(6, 8), (4, 2)],
    8: [(6, 8), (4, 2)],
    9: [(8, 8), (3, 3)],
    10: [(12, 8), (4, 3)],
    11: [(12, 8), (4, 3)],
    12: [(12, 8), (4, 3)],
}


def plot_irfs(
    sims,
    variables=None,
    titles=None,
    layout=None,
    horizon=None,
    figsize=None,
    plot_options={},
    line_options=None,
):

    from matplotlib import pyplot as plt

    if not isinstance(sims, list):
        # we have only one series to compare
        sims = [sims]

    if variables is None:
        variables = sims[0].columns

    if horizon is None:
        horizon = sims[0].shape[0]

    if titles is None:
        titles = variables

    n_v = len(variables)
    max_v = max(default_layout.keys())
    if n_v > max_v:
        raise Exception("Too many graphs")

    if layout is None:
        if figsize is None:
            figsize, layout = default_layout[n_v]
        else:
            layout = default_layout[n_v][1]

    if line_options is None:
        line_options = [{}] * n_v

    fig, axes = plt.subplots(layout[0], layout[1], figsize=figsize)
    axes = axes.ravel()
    for i, v in enumerate(variables):
        ax = axes[i]
        for j, sim in enumerate(sims):
            arguments = {}
            arguments.update(plot_options)
            arguments.update(line_options[j])
            ax.plot(sim[v][:horizon], **arguments)
        ax.set_title(titles[i])
        ax.set_xlim(0, horizon)
        max_ = max([sim[v].max() for sim in sims])
        min_ = min([sim[v].min() for sim in sims])
        if (max_ - min_) > 0.001:  # pretty arbitrary
            margin = (max_ - min_) * 0.05  # 5% on each side
            ax.set_ylim(min_ - margin, max_ + margin)
        ax.grid(True)
    # fig.tight_layout()

    return fig


if __name__ == "__main__":

    A = np.random.random((50, N))
    A *= 0.9 ** (np.arange(N))[None, :]

    B = np.random.random((50, N))
    B *= 0.9 ** (np.arange(N))[None, :]

    from matplotlib import pyplot

    from string import ascii_lowercase
    from random import choice

    letters = [choice(ascii_lowercase) for _ in range(N)]

    df1 = pd.DataFrame(A, columns=letters)
    df2 = pd.DataFrame(B, columns=letters)



# def test_lint_all():

#     from dolo import yaml_import
#     import os
#     import re

#     examples_dir = 'examples/models/'
#     examples = os.listdir(examples_dir)

#     regex = re.compile("(.*).yaml")

#     for f in examples:
#         if not regex.match(f):
#             continue
#         print("Checking: {}".format(f))
#         check = yaml_import(examples_dir + f, check_only=True)

#         # TODO: need to fix Dynare lint before reenabling
#         # if len(check)>1:
#         #     assert("Linter Error" not in check[0]) # assert there is no linter error
#         print()

# if __name__ == "__main__":
#     test_lint_all()



# from ast import *
from dolang.symbolic import stringify, stringify_symbol, parse_string, list_variables
from dolang.grammar import str_expression

from dolo.compiler.misc import CalibrationDict
from numpy import log, exp

import xarray


def eval_formula(expr: str, dataframe=None, context=None):
    """
    expr: string
        Symbolic expression to evaluate.
        Example: `k(1)-delta*k(0)-i`
    table: (optional) pandas dataframe
        Each column is a time series, which can be indexed with dolo notations.
    context: dict or CalibrationDict
    """

    if context is None:
        dd = {}  # context dictionary
    elif isinstance(context, CalibrationDict):
        dd = context.flat.copy()
    else:
        dd = context.copy()

    # compat since normalize form for parameters doesn't match calib dict.
    for k in [*dd.keys()]:
        dd[stringify_symbol(k)] = dd[k]

    expr_ast = parse_string(expr)
    variables = list_variables(expr_ast)
    nexpr = stringify(expr_ast)

    dd["log"] = log
    dd["exp"] = exp

    if dataframe is not None:

        import pandas as pd

        for (k, t) in variables:
            dd[stringify_symbol((k, t))] = dataframe[k].shift(t)
        dd["t_"] = pd.Series(dataframe.index, index=dataframe.index)

    expr = str_expression(nexpr)

    res = eval(expr.replace("^", "**"), dd)

    return res



import functools


class memoized(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kargs):

        targs = (e for e in args)
        hh = tuple(hashable(e) for e in targs)
        h2 = hashable(kargs)
        h = hash((hh, h2))
        try:
            return self.cache[h]
        except KeyError:
            value = self.func(*args, **kargs)
            self.cache[h] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args, **kargs)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


class cachedondisk(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """

    def __init__(self, func):

        # create caching direcory if it is not there already
        import os

        if not os.path.isdir(".cache"):
            os.mkdir(".cache")

        self.func = func
        self.fname = func.__name__

    def __call__(self, *args, **kargs):
        import pickle

        hh = tuple(hashable(e) for e in args)
        h2 = hashable(kargs)
        h = hash((hh, h2))
        try:
            with open(".cache/{0}.{1}.pickle".format(self.fname, h), "rb") as f:
                value = pickle.load(f)
            return value
        except IOError:
            value = self.func(*args, **kargs)
            if value is not None:  # should there be other kinds of error values
                # write file with h
                with open(".cache/{0}.{1}.pickle".format(self.fname, h), "wb") as f:
                    pickle.dump(value, f)
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args, **kargs)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


def clear_cache(function_name=None):

    import os

    try:
        if function_name:
            os.system("rm -rf .cache/{}.*.pickle".format(function_name))
        else:
            os.system("rm -rf .cache/*.pickle")
    except:
        pass


import os


class DiskDictionary:
    def __init__(self, directory=".cache", funname="fun"):
        self.funname = funname
        if not os.path.isdir(directory):
            os.mkdir(directory)
        self.directory = directory

    def get_filename(self, key):
        import pickle

        hh = tuple(hashable(k) for k in key)
        h = hash(hh)
        filename = "{0}/{1}.{2}.pickle".format(self.directory, self.funname, h)
        return filename

    def __setitem__(self, key, value):
        import pickle

        filename = self.get_filename(key)
        try:
            with open(filename, "w") as f:
                pickle.dump(value, f)
        except TypeError as e:
            raise e

    def get(self, item):
        import pickle

        filename = self.get_filename(item)
        try:
            with open(filename) as f:
                value = pickle.load(f)
                return value
        except:
            return None


import collections


def hashable(obj):
    if hasattr(obj, "flatten"):  # for numpy arrays
        return tuple(obj.flatten().tolist())
    if isinstance(obj, collections.abc.Hashable):
        return obj
    if isinstance(obj, collections.abc.Mapping):
        items = [(k, hashable(v)) for (k, v) in obj.items()]
        return frozenset(items)
    if isinstance(obj, collections.abc.Iterable):
        return tuple([hashable(item) for item in obj])
    return TypeError(type(obj))



import unittest

import numpy as np


def josephy(x):
    #   Computes the function value F(x) of the NCP-example by Josephy.
    n = len(x)
    Fx = np.zeros(n)
    Fx[0] = 3 * x[0] ** 2 + 2 * x[0] * x[1] + 2 * x[1] ** 2 + x[2] + 3 * x[3] - 6
    Fx[1] = 2 * x[0] ** 2 + x[0] + x[1] ** 2 + 3 * x[2] + 2 * x[3] - 2
    Fx[2] = 3 * x[0] ** 2 + x[0] * x[1] + 2 * x[1] ** 2 + 2 * x[2] + 3 * x[3] - 1
    Fx[3] = x[0] ** 2 + 3 * x[1] ** 2 + 2 * x[2] + 3 * x[3] - 3
    return Fx


def Djosephy(x):
    # Local Variables: x, DFx, n
    # Function calls: Djosephy, zeros, length
    #%
    #%   Computes the Jacobian DF(x) of the NCP-example by Josephy
    #%
    n = len(x)
    DFx = np.zeros((n, n))
    DFx[0, 0] = 6.0 * x[0] + 2.0 * x[1]
    DFx[0, 1] = 2.0 * x[0] + 4.0 * x[1]
    DFx[0, 2] = 1.0
    DFx[0, 3] = 3.0
    DFx[1, 0] = 4.0 * x[0] + 1.0
    DFx[1, 1] = 2.0 * x[1]
    DFx[1, 2] = 3.0
    DFx[1, 3] = 2.0
    DFx[2, 0] = 6.0 * x[0] + x[1]
    DFx[2, 1] = x[0] + 4.0 * x[1]
    DFx[2, 2] = 2.0
    DFx[2, 3] = 3.0
    DFx[3, 0] = 2.0 * x[0]
    DFx[3, 1] = 6.0 * x[1]
    DFx[3, 2] = 2.0
    DFx[3, 3] = 3.0
    return DFx


class ComplementaritySolve(unittest.TestCase):

    #
    # # TODO: the two first tests fail because ncpsolve assumes a vectorize function
    # def test_infinite_bounds(self):
    #
    #     import numpy
    #     f = lambda x: [-x**3 + 1.2, -numpy.atleast_2d(3*x**2)]
    #     lb = numpy.array([-numpy.inf])
    #     ub = numpy.array([numpy.inf])
    #     x0 = numpy.array([0.3])
    #     res = ncpsolve(f, lb, ub, x0)
    #
    # def test_complementarities(self):
    #
    #     import numpy
    #     from numpy.testing import assert_almost_equal
    #     f = lambda x: [-x**3 + 1.2, -numpy.atleast_2d(3*x**2)]
    #     lb = numpy.array([-1])
    #     ub = numpy.array([1])
    #     x0 = numpy.array([0.3])
    #     res = ncpsolve(f, lb, ub, x0)
    #     assert_almost_equal( res, 1.0)
    #
    # def test_josephy(self):
    #
    #     import numpy
    #
    #     fun = lambda x: [-josephy(x), -Djosephy(x)]
    #
    #     x0 = np.array( [1.25, 0.01, 0.01, 0.50] )
    #
    #     lb = np.array( [0.00, 0.00, 0.00, 0.00] )
    #     ub = np.array( [inf, inf, inf, inf] )
    #
    #     resp = ncpsolve(fun,  lb, ub, x0, tol=1e-15)
    #
    #     sol = numpy.array( [ 1.22474487e+00, 0.00000000e+00, 3.60543164e-17, 5.00000000e-01])
    #
    #     from numpy.testing import assert_almost_equal
    #
    #     assert_almost_equal(sol,  resp)

    def test_lmmcp(self):

        from dolo.numeric.extern import lmmcp

        x0 = np.array([1.25, 0.00, 0.00, 0.50])
        lb = np.array([0.00, 0.00, 0.00, 0.00])
        ub = np.array([1e20, 1e20, 1e20, 1e20])

        resp = lmmcp.lmmcp(josephy, Djosephy, x0, lb, ub)

        sol = np.array([1.22474487, -0.0000, 0.0000, 0.5000])
        print(sol)
        print(resp)
        assert abs(sol - resp).max() < 1e-5


if __name__ == "__main__":

    unittest.main()



from dolo.version import __version_info__, __version__

import dolang

from dolo.config import *

import dolo.compiler.objects
import dolo.numeric.processes
import dolo.numeric.distribution

# import dolo.numeric.grids
# del dolo.compiler.objects
# del dolo.numeric.processes
# del dolo.numeric.distribution
# del dolo.numeric.grids

from dolo.compiler.model_import import yaml_import
from dolo.misc.display import pcat
from dolo.misc.groot import groot
from dolo.misc.dprint import dprint

from dolo.algos.commands import *
